/*****************************************************************************

        RectalAnarchy.cpp
        By Laurent de Soras, 2001

--- Legal stuff ---

This program is free software. It comes without any warranty, to
the extent permitted by applicable law. You can redistribute it
and/or modify it under the terms of the Do What The Fuck You Want
To Public License, Version 2, as published by Sam Hocevar. See
http://sam.zoy.org/wtfpl/COPYING for more details.

*Tab=3***********************************************************************/



#if defined (_MSC_VER)
	#pragma warning (1 : 4130 4223 4705 4706)
	#pragma warning (4 : 4355 4786 4800)
#endif



/*\\\ INCLUDE FILES \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/

#include	"basic/basic_def.h"
#include	"basic/basic_fnc.h"
#include	"RectalAnarchy.h"

#include	<cassert>
#include	<cstdio>
#include	<cmath>



namespace std {};



/*\\\ PUBLIC \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/



extern "C"
{

__declspec (dllexport) CMachineInfo const * __cdecl GetInfo ()
{
	return (&RectalAnarchy::_mac_info);
}

__declspec (dllexport) CMachineInterface * __cdecl CreateMachine ()
{
	return (new RectalAnarchy);
}

}



const CMachineInfo	RectalAnarchy::_mac_info =
{
	MT_GENERATOR,							// type, MT_GENERATOR or MT_EFFECT
	MI_VERSION,								// version, always set this to MI_VERSION
	0,											// Flags ????
	MIN_NBR_TRACKS,						// minimum number of tracks
	MAX_NBR_TRACKS,						// maximum number of tracks
	NBR_GLOB_PARAM,						// number of global parameters
	NBR_TRACK_PARAM,						// number of track parameters
	_param_list,							// pointer to list of parameters
	NBR_ATTRIBUTES,						// number of attributes
	_attrib_list,							// pointer to list of attributes
#if ! defined (NDEBUG)
	"RectalAnarchy (debug build)",	// name for debug build
#else
	"RectalAnarchy",						// name for release build
#endif
	"RectalAnarchy",						// short name
	"FireSledge",							// your name
	NULL										// list of commands
};



RectalAnarchy::RectalAnarchy ()
:	_nbr_tracks (-1)
,	_nbr_channels (-1)
{
	assert (NBR_GLOB_PARAM >= 0);
	assert (NBR_TRACK_PARAM >= 0);
	assert (NBR_ATTRIBUTES >= 0);
	assert (MIN_NBR_TRACKS >= 0);
	assert (MAX_NBR_TRACKS >= MIN_NBR_TRACKS);
	assert (MAX_POLYPHONY >= 0);

	GlobalVals = &_gval;
	TrackVals = _tval;
	AttrVals = reinterpret_cast <int *> (&_aval);
}



RectalAnarchy::~RectalAnarchy ()
{
}



int	RectalAnarchy::get_polyphony ()
{
	assert (_nbr_channels >= 0);

	return (_nbr_channels);
}



RectalAnarchyChannel &	RectalAnarchy::use_channel (int chn)
{
	assert (chn >= 0);
	assert (chn < _nbr_channels);
	assert (chn < MAX_POLYPHONY);

	return (_channels [chn]);
}



float *	RectalAnarchy::use_adsr_temp_buf ()
{
	return (_adsr_temp_buf);
}



const RectalAnarchy::Formant &	RectalAnarchy::get_formant_data (int fmt) const
{
	assert (fmt >= 0);
	assert (fmt < RectalAnarchyChannel::NBR_FORMANTS);

	return (_formants [fmt]);
}



void	RectalAnarchy::Init (CMachineDataInput * const pi)
{
	int				track;

	_sample_freq = float (pMasterInfo->SamplesPerSec);
	assert (_sample_freq > 0);
	assert (_param_list [NBR_GLOB_PARAM + NBR_TRACK_PARAM - 1] != 0);
	assert (_attrib_list [NBR_ATTRIBUTES - 1] != 0);

	_midi_channel = -1;
	_cur_midi_track = 0;
	_nbr_tracks = MIN_NBR_TRACKS;
	_nbr_channels = _attrib_polyphony.DefValue;

	for (int fmt = 0; fmt < RectalAnarchyChannel::NBR_FORMANTS; ++fmt)
	{
		_formants [fmt]._filter.set_sample_freq (_sample_freq);
	}

	for (int chn = 0; chn < MAX_POLYPHONY; ++chn)
	{
		_channels [chn].init (*this, _sample_freq);
	}

	_voice_type.set_sample_freq (_sample_freq);
	_voice_sound.set_sample_freq (_sample_freq);
	_voice_lfo.set_sample_freq (_sample_freq);

	for (track = 0; track < MAX_NBR_TRACKS; ++track)
	{
		_tracks [track].init (*this, _sample_freq);
	}

	// Default settings
	set_glide_time (buzz_2_glide_time (_param_glide_time.DefValue));
	set_voice_type (buzz_2_voice_type (_param_voice_type.DefValue));
	set_voice_sound (buzz_2_voice_sound (_param_voice_sound.DefValue));
	set_voice_lfo_depth (buzz_2_voice_lfo_depth (_param_voice_lfo_depth.DefValue));
	set_voice_lfo_per (buzz_2_voice_lfo_per (_param_voice_lfo_per.DefValue));
	set_voice_lfo_type (buzz_2_voice_lfo_type (_param_voice_lfo_type.DefValue));
	set_emphasis (buzz_2_emphasis (_param_emphasis.DefValue));
	set_vca_at (buzz_2_vca_at (_param_vca_at.DefValue));
	set_vca_dt (buzz_2_vca_dt (_param_vca_dt.DefValue));
	set_vca_sl (buzz_2_vca_sl (_param_vca_sl.DefValue));
	set_vca_st (buzz_2_vca_st (_param_vca_st.DefValue));
	set_vca_rt (buzz_2_vca_rt (_param_vca_rt.DefValue));

	for (track = 0; track < MAX_NBR_TRACKS; ++track)
	{
		_tracks [track].set_velo (buzz_2_velo (_param_velo.DefValue));
	}

	// Load values
	if (pi != NULL)
	{
		byte			version;
		byte			nbr_tracks;

		pi->Read (version);
		pi->Read (nbr_tracks);

		GlobalParameters	glob_val;
		pi->Read (&glob_val, sizeof (glob_val));

		set_glide_time (glob_val._glide_time);
		set_voice_type (glob_val._voice_type);
		set_voice_sound (glob_val._voice_sound);
		set_voice_lfo_depth (glob_val._voice_lfo_depth);
		set_voice_lfo_per (glob_val._voice_lfo_per);
		set_voice_lfo_type (glob_val._voice_lfo_type);
		set_emphasis (glob_val._emphasis);
		set_vca_at (glob_val._vca_at);
		set_vca_dt (glob_val._vca_dt);
		set_vca_sl (glob_val._vca_sl);
		set_vca_st (glob_val._vca_st);
		set_vca_rt (glob_val._vca_rt);

		for (track = 0; track < nbr_tracks; ++track)
		{
			_tracks [track].load (pi);
		}
	}

	_voice_type.clear_buffers ();
	_voice_sound.clear_buffers ();
}



void	RectalAnarchy::Tick ()
{
	// Global parameters
	if (_gval._glide_time != _param_glide_time.NoValue)
	{
		set_glide_time (buzz_2_glide_time (_gval._glide_time));
	}
	if (_gval._voice_type != _param_voice_type.NoValue)
	{
		set_voice_type (buzz_2_voice_type (_gval._voice_type));
	}
	if (_gval._voice_sound != _param_voice_sound.NoValue)
	{
		set_voice_sound (buzz_2_voice_sound (_gval._voice_sound));
	}
	if (_gval._voice_lfo_depth != _param_voice_lfo_depth.NoValue)
	{
		set_voice_lfo_depth (buzz_2_voice_lfo_depth (_gval._voice_lfo_depth));
	}
	if (_gval._voice_lfo_per != _param_voice_lfo_per.NoValue)
	{
		set_voice_lfo_per (buzz_2_voice_lfo_per (_gval._voice_lfo_per));
	}
	if (_gval._voice_lfo_type != _param_voice_lfo_type.NoValue)
	{
		set_voice_lfo_type (buzz_2_voice_lfo_type (_gval._voice_lfo_type));
	}
	if (_gval._emphasis != _param_emphasis.NoValue)
	{
		set_emphasis (buzz_2_emphasis (_gval._emphasis));
	}
	if (_gval._glottal_open_time != _param_glottal_open_time.NoValue)
	{
		set_glottal_open_time (buzz_2_glottal_open_time (_gval._glottal_open_time));
	}
	if (_gval._vca_at != _param_vca_at.NoValue)
	{
		set_vca_at (buzz_2_vca_at (_gval._vca_at));
	}
	if (_gval._vca_dt != _param_vca_dt.NoValue)
	{
		set_vca_dt (buzz_2_vca_dt (_gval._vca_dt));
	}
	if (_gval._vca_sl != _param_vca_sl.NoValue)
	{
		set_vca_sl (buzz_2_vca_sl (_gval._vca_sl));
	}
	if (_gval._vca_st != _param_vca_st.NoValue)
	{
		set_vca_st (buzz_2_vca_st (_gval._vca_st));
	}
	if (_gval._vca_rt != _param_vca_rt.NoValue)
	{
		set_vca_rt (buzz_2_vca_rt (_gval._vca_rt));
	}

	// Track parameters
	for (int track = 0; track < _nbr_tracks; ++track)
	{
		float				slide_time = -1;

		// To be efficient, slide speed must be taken in account before note
		// parameter
		if (_tval [track]._slide != _param_slide.NoValue)
		{
			slide_time = buzz_2_slide (_tval [track]._slide);
		}

		if (_tval [track]._note != _param_note.NoValue)
		{
			const int		note = _tval [track]._note;
			if (note == NOTE_OFF)
			{
				_tracks [track].note_off ();
			}
			else
			{
				const float		val = buzz_2_note (_tval [track]._note);
				_tracks [track].note_on (val, slide_time);
			}
		}

		if (_tval [track]._velo != _param_velo.NoValue)
		{
			const float		val = buzz_2_velo (_tval [track]._velo);
			_tracks [track].set_velo (val);
		}
	}
}



bool	RectalAnarchy::Work (float *psamples, int numsamples, int const mode)
{
	bool				gen_flag = false;

	if (   _voice_type.is_ramping ()
	    || _voice_sound.is_ramping ())
	{
		_need_formant_update_flag = true;
	}
	_voice_type.tick (numsamples);
	_voice_sound.tick (numsamples);

	if (_voice_lfo.get_amp () > 0)
	{
		_voice_lfo.tick (numsamples);
		_need_formant_update_flag = true;
	}

	if ((mode & WM_WRITE) == 0)
	{
		// Off or Bypass with silent input
		if ((mode & WM_READ) == 0)
		{
			gen_flag = true;
		}

		// Bypass with input
		else
		{
			gen_flag = false;
		}
	}

	else
	{
		if (_need_formant_update_flag)
		{
			setup_formant_filters ();
		}

		gen_flag = false;
		for (int chn = 0; chn < _nbr_channels; ++chn)
		{
			gen_flag |= _channels [chn].generate (psamples, numsamples, gen_flag);
		}
	}

	return (gen_flag);
}



void	RectalAnarchy::Stop ()
{
	stop_all_notes ();
	stop_all_channels ();
}



void	RectalAnarchy::Save (CMachineDataOutput * const po)
{
	assert (po != 0);

	po->Write ((byte) MCH_VERSION);
	po->Write ((byte) _nbr_tracks);

	po->Write (&_global_param, sizeof (_global_param));
	for (int track = 0; track < _nbr_tracks; ++track)
	{
		_tracks [track].save (po);
	}
}



void	RectalAnarchy::AttributesChanged ()
{
	const int		new_sound_quality = _aval._sound_quality;

	/*** A finir ***/

	const int		new_poly = _aval._polyphony;
	if (new_poly != _nbr_channels)
	{
		stop_all_notes ();
		stop_all_channels ();
		_nbr_channels = new_poly;
	}

	const int		new_midi_chn = _aval._midi_channel - 1;
	if (new_midi_chn != _midi_channel)
	{
		_midi_channel = new_midi_chn;
	}
}



void	RectalAnarchy::SetNumTracks (int const n)
{
	int				track;

	// Track reduction
	for (track = n; track < _nbr_tracks; ++track)
	{
		_tracks [track].note_off ();
	}

	_nbr_tracks = n;
	_cur_midi_track = 0;
}



void	RectalAnarchy::MuteTrack (int const i)
{

	/*** A faire ***/

}



void	RectalAnarchy::MidiNote (int const channel, int const value, int const velocity)
{
	if (   _midi_channel < 0
	    || _midi_channel == channel)
	{
		int				cur_track = _cur_midi_track;

		// Note Off
		if (velocity <= 0)
		{
			bool				found_flag = false;
			for (int cnt = 0; cnt < _nbr_tracks; ++cnt)
			{
				if (_tracks [cur_track].get_midi_note () == value)
				{
					found_flag = true;
					break;
				}
				cur_track = (cur_track + _nbr_tracks - 1) % _nbr_tracks;
			}

			if (found_flag)
			{
				_tracks [cur_track].note_off ();
			}
		}

		// Note On
		else
		{
			bool				found_flag = false;
			for (int cnt = 0; cnt < _nbr_tracks; ++cnt)
			{
				if (_tracks [cur_track].get_midi_note () < 0)
				{
					found_flag = true;
					break;
				}
				cur_track = (cur_track + 1) % _nbr_tracks;
			}

			if (found_flag)
			{
				_tracks [cur_track].set_velo (buzz_2_velo (velocity));
				_tracks [cur_track].note_on (float (value), 0);
				_cur_midi_track = (cur_track + 1) % _nbr_tracks;
			}
		}
	}
}



void	RectalAnarchy::Event (dword const data)
{

	/*** A faire ***/

}



const char *	RectalAnarchy::DescribeValue (int const param, int const value)
{
	using namespace std;

	switch (param)
	{
	// Glide time
	case	0:
		sprintf (_txt_0, "%.1lf ticks", double (buzz_2_glide_time (value)));
		break;

	// Voice type
	case	1:
		{
			const double	v_index = buzz_2_voice_type (value);
			const int		v_index_0 = basic::floor_int (v_index);
			const int		v_index_1 = (v_index_0 + 1) % NBR_TRANSITIONS;
			const int		v_0 = TRANSITION_LIST [v_index_0];
			const int		v_1 = TRANSITION_LIST [v_index_1];
			const double	v_frac = v_index - v_index_0;

			const int		interval = 16;
			const int		length = 12;
			assert (length <= interval - 1);
			char				raw_txt [interval * 2];
			memset (raw_txt, '-', sizeof (raw_txt));
			const char *	n0_0 = VOICE_LIST [v_0]._name_0;
			const char *	n1_0 = VOICE_LIST [v_1]._name_0;
			memcpy (raw_txt,
			        n0_0,
					  basic::min (int (strlen (n0_0)), interval));
			memcpy (raw_txt + interval,
			        n1_0,
			        basic::min (int (strlen (n1_0)), interval));
			memcpy (_txt_0,
			        raw_txt + basic::round_int (v_frac * interval),
			        length);
			_txt_0 [length] = '\0';
		}
		break;

	// Voice sound
	case	2:
		{
			const double	v_index = buzz_2_voice_sound (value);
			const int		v_index_0 = basic::floor_int (v_index);
			const int		v_index_1 = (v_index_0 + 1) % NBR_TRANSITIONS;
			const int		v_0 = TRANSITION_LIST [v_index_0];
			const int		v_1 = TRANSITION_LIST [v_index_1];
			const double	v_frac = v_index - v_index_0;

			const int		interval = 8;
			const int		length = 8;
			assert (length <= interval);
			char				raw_txt [interval * 2];
			memset (raw_txt, '-', sizeof (raw_txt));
			const char *	n0_0 = VOWEL_NAME_0 [v_0];
			const char *	n1_0 = VOWEL_NAME_0 [v_1];
			memcpy (raw_txt,
			        n0_0,
					  basic::min (int (strlen (n0_0)), interval));
			memcpy (raw_txt + interval,
			        n1_0,
			        basic::min (int (strlen (n1_0)), interval));
			memcpy (_txt_0,
			        raw_txt + basic::round_int (v_frac * interval),
			        length);
			_txt_0 [length] = '\0';
		}
		break;

	// Voice sound LFO depth
	case	3:
		sprintf (_txt_0, "%.1lf %%", double (buzz_2_voice_lfo_depth (value) * 100));
		break;

	// Voice sound LFO period
	case	4:
		sprintf (_txt_0, "%.1lf ticks", double (buzz_2_voice_lfo_per (value)));
		break;

	// Voice sound LFO waveform
	case	5:
		sprintf (_txt_0, "%s", dsp::Lfo::WF_TYPE_NAME_0 [buzz_2_voice_lfo_type (value)]);
		break;

	// Emphasis
	case	6:
		sprintf (_txt_0, "%d", value);
		break;

	// Glottal opening time
	case	7:
		sprintf (_txt_0, "%.1lf ms", double (buzz_2_glottal_open_time (value) * 1000));
		break;

	// VCA attack time
	case	8:
		sprintf (_txt_0, "%.1lf ms", double (buzz_2_vca_at (value) * 1000));
		break;

	// VCA decay time
	case	9:
		sprintf (_txt_0, "%.1lf ms", double (buzz_2_vca_dt (value) * 1000));
		break;

	// VCA sustain level
	case	10:
		sprintf (_txt_0, "%.1lf %%", double (buzz_2_vca_sl (value) * 100));
		break;

	// VCA sustain time
	case	11:
		{
			const double	val = buzz_2_vca_st (value);
			if (val >= dsp::AdsrBase::INFINITE_SUSTAIN - 1)
			{
				strcpy (_txt_0, "Infinite");
			}
			else
			{
				sprintf (_txt_0, "%.1lf ms", double (val * 1000));
			}
		}
		break;

	// VCA release time
	case	12:
		sprintf (_txt_0, "%.1lf ms", double (buzz_2_vca_rt (value) * 1000));
		break;

	// Note
	case	13:
		sprintf (_txt_0, "%d", value);
		break;

	// Velocity
	case	14:
		sprintf (_txt_0, "%d", value);
		break;

	// Slide time
	case	15:
		sprintf (_txt_0, "%.1lf ticks", double (buzz_2_slide (value)));
		break;

	default:
		sprintf (_txt_0, "%d", value);
		assert (false);
		break;

	}

	return (_txt_0);
}



/*\\\ PROTECTED \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/



/*\\\ PRIVATE \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/



float	RectalAnarchy::buzz_2_note (int param) const
{
	return (float (((param >> 4) & 0xF) * 12 + (param & 0xF) - 1));
}



float	RectalAnarchy::buzz_2_velo (int param) const
{
	const float		val = float (param) / _param_velo.DefValue;

	return (val);
}



float	RectalAnarchy::buzz_2_slide (int param) const
{
	return (float (param) / 0x04);
}



float	RectalAnarchy::buzz_2_glide_time (int param) const
{
	assert (param >= 0);

	return (float (param) / 0x100);
}



float	RectalAnarchy::buzz_2_voice_type (int param) const
{
	const int		int_part = (param >> 12) & 0xF;
	const float		frac_part = (param & 0x0FFF) * (1.0f / 0x1000);
	assert (int_part >= 0);
	assert (int_part < NBR_TRANSITIONS);
	assert (frac_part >= 0);
	assert (frac_part < 1);

	return (int_part + frac_part);
}



float	RectalAnarchy::buzz_2_voice_sound (int param) const
{
	return (buzz_2_voice_type (param));
}



float	RectalAnarchy::buzz_2_voice_lfo_depth (int param) const
{
	assert (param >= 0);

	const double	ratio = double (param) / _param_voice_lfo_depth.MaxValue;
	const double	val = basic::pseudo_exp10 (ratio, 1);

	return (float (val));
}



float	RectalAnarchy::buzz_2_voice_lfo_per (int param) const
{
	assert (param >= 0);

	const double	ratio = double (param) / _param_voice_lfo_per.MaxValue;
	const double	val = exp ((ratio * 14 - 2) * basic::LN2);

	return (float (val));
}



dsp::Lfo::WfType	RectalAnarchy::buzz_2_voice_lfo_type (int param) const
{
	assert (param >= 0);
	assert (param < NBR_LFO);

	static const dsp::Lfo::WfType	table [NBR_LFO] =
	{
		dsp::Lfo::SINE,
		dsp::Lfo::TRIANGLE,
		dsp::Lfo::SQUARE,
		dsp::Lfo::RAMP_UP,
		dsp::Lfo::RAMP_DOWN,
		dsp::Lfo::BROWN_NOISE
	};

	return (table [param]);
}



float	RectalAnarchy::buzz_2_emphasis (int param) const
{
	using namespace std;

	const double	ratio = double (param) / _param_emphasis.MaxValue;
	const double	val = exp (ratio * basic::LN10);

	return (float (val));
}



float	RectalAnarchy::buzz_2_glottal_open_time (int param) const
{
	using namespace std;

	const double	ratio = double (param) / _param_glottal_open_time.MaxValue;
	const double	val = exp ((ratio - 3) * basic::LN10);

	return (float (val));
}



float	RectalAnarchy::buzz_2_vca_at (int param) const
{
	using namespace std;

	const double	ratio = double (param) / _param_vca_at.MaxValue;
	const double	val = exp ((ratio * 5 - 3) * basic::LN10);

	return (float (val));
}



float	RectalAnarchy::buzz_2_vca_dt (int param) const
{
	using namespace std;

	const double	ratio = double (param) / _param_vca_dt.MaxValue;
	const double	val = exp ((ratio * 5 - 3) * basic::LN10);

	return (float (val));
}



float	RectalAnarchy::buzz_2_vca_sl (int param) const
{
	return (float (param) / _param_vca_sl.MaxValue);
}



float	RectalAnarchy::buzz_2_vca_st (int param) const
{
	using namespace std;

	if (param == _param_vca_st.MinValue)
	{
		return (dsp::AdsrBase::INFINITE_SUSTAIN);
	}
	const double	ratio = double (param) / _param_vca_st.MaxValue;
	const double	val = exp ((ratio * 5 - 3) * basic::LN10);

	return (float (val));
}



float	RectalAnarchy::buzz_2_vca_rt (int param) const
{
	using namespace std;

	const double	ratio = double (param) / _param_vca_rt.MaxValue;
	const double	val = exp ((ratio * 5 - 3) * basic::LN10);

	return (float (val));
}



void	RectalAnarchy::set_glide_time (float t)
{
	assert (t >= 0);
	_global_param._glide_time = t;

	const float			val_s = t / pMasterInfo->TicksPerSec;
	_voice_type.set_inertia_time (val_s);
	_voice_sound.set_inertia_time (val_s);
}



void	RectalAnarchy::set_voice_type (float val)
{
	_global_param._voice_type = val;
	_voice_type.set_val (val);
	_need_formant_update_flag = true;
}



void	RectalAnarchy::set_voice_sound (float val)
{
	_global_param._voice_sound = val;
	_voice_sound.set_val (val);
	_need_formant_update_flag = true;
}



void	RectalAnarchy::set_voice_lfo_depth (float val)
{
	_global_param._voice_lfo_depth = val;
	_voice_lfo.set_amp (val);
	_need_formant_update_flag = true;	// We need that if we turn the LFO off.
}



void	RectalAnarchy::set_voice_lfo_per (float val)
{
	_global_param._voice_lfo_per = val;

	const float			val_s = val / pMasterInfo->TicksPerSec;
	_voice_lfo.set_period (val_s);
}



void	RectalAnarchy::set_voice_lfo_type (dsp::Lfo::WfType val)
{
	_global_param._voice_lfo_type = val;
	_voice_lfo.set_type (val);
}



void	RectalAnarchy::set_emphasis (float val)
{
	using namespace std;

	_global_param._emphasis = val;
	_filter_gain = float (1.0 / sqrt (val));	// Gain compensation. Seems to be quite good
	_need_formant_update_flag = true;
}



void	RectalAnarchy::set_glottal_open_time (float t)
{
	_global_param._glottal_open_time = t;
	for (int chn = 0; chn < MAX_POLYPHONY; ++chn)
	{
		_channels [chn].set_glottal_open_time (t);
	}
}



void	RectalAnarchy::set_vca_at (float t)
{
	_global_param._vca_at = t;
	for (int chn = 0; chn < MAX_POLYPHONY; ++chn)
	{
		_channels [chn].set_vca_at (t);
	}
}



void	RectalAnarchy::set_vca_dt (float t)
{
	_global_param._vca_dt = t;
	for (int chn = 0; chn < MAX_POLYPHONY; ++chn)
	{
		_channels [chn].set_vca_dt (t);
	}
}



void	RectalAnarchy::set_vca_sl (float lvl)
{
	_global_param._vca_sl = lvl;
	for (int chn = 0; chn < MAX_POLYPHONY; ++chn)
	{
		_channels [chn].set_vca_sl (lvl);
	}
}



void	RectalAnarchy::set_vca_st (float t)
{
	_global_param._vca_st = t;
	for (int chn = 0; chn < MAX_POLYPHONY; ++chn)
	{
		_channels [chn].set_vca_st (t);
	}
}



void	RectalAnarchy::set_vca_rt (float t)
{
	_global_param._vca_rt = t;
	for (int chn = 0; chn < MAX_POLYPHONY; ++chn)
	{
		_channels [chn].set_vca_rt (t);
	}
}



void	RectalAnarchy::setup_formant_filters ()
{
	using namespace std;

	float				vowel_val = _voice_sound.get_val ();
	if (_voice_lfo.get_amp () > 0)
	{
		vowel_val += _voice_lfo.get_value () * NBR_TRANSITIONS / 2;
		while (vowel_val < 0)
		{
			vowel_val += NBR_TRANSITIONS;
		}
		while (vowel_val >= NBR_TRANSITIONS)
		{
			vowel_val -= NBR_TRANSITIONS;
		}
	}

	const int		vowel_int = basic::floor_int (vowel_val);
	const float		vowel_frac = vowel_val - vowel_int;
	const int		vowel_0 = TRANSITION_LIST [vowel_int];
	const int		vowel_1 = TRANSITION_LIST [(vowel_int + 1) % NBR_TRANSITIONS];

	const float		voice_val = _voice_type.get_val ();
	const int		voice_int = basic::floor_int (voice_val);
	const float		voice_frac = voice_val - voice_int;
	const int		voice_0 = TRANSITION_LIST [voice_int];
	const int		voice_1 = TRANSITION_LIST [(voice_int + 1) % NBR_TRANSITIONS];

	// Sound data to interpolate
	const VowelData * const	vowel_data_ptr [2] [2] =
	{
		{
			&VOICE_LIST [voice_0]._vowel [vowel_0],
			&VOICE_LIST [voice_0]._vowel [vowel_1]
		},
		{
			&VOICE_LIST [voice_1]._vowel [vowel_0],
			&VOICE_LIST [voice_1]._vowel [vowel_1]
		}
	};

	// Coefficients for bilinear interpolation
	float				coef [2] [2];
	coef [1] [1] = voice_frac * vowel_frac;
	coef [1] [0] = voice_frac - coef [1] [1];
	coef [0] [1] = vowel_frac - coef [1] [1];
	coef [0] [0] = 1 - coef [1] [1] - coef [0] [1] - coef [1] [0];

	for (int fmt = 0; fmt < RectalAnarchyChannel::NBR_FORMANTS; ++fmt)
	{
//1000 0 497 0  739 0 2772 0 3364 0 4170 0 4000 0 0 0 200 40  0 40  0 20  0  0 53 44 79 70 52 95 44 56 34 80  0 80  0  0 27 70
// 800 0 450 0 1556 0 1963 0 3088 0 4249 0 4000 0 0 0 200 40  0 40  0 20  0  0 54 51 65 51 60 50 49 49 45 80  0 80  0  0 60 70
//		static const float	fmt_freq [] = { 450, 1556, 1963, 3088, 4249 };
//		static const float	fmt_bw [] = { 51, 51, 50, 49, 80 };
//		static const float	fmt_gain [] = { 2.0f, 6.0f, 3.0f, 1.0f, 0.5f };

		// Bilinear interpolation of formant parameters
		_formants [fmt]._center_freq =
			  vowel_data_ptr [0] [0]->_freq [fmt] * coef [0] [0]
			+ vowel_data_ptr [0] [1]->_freq [fmt] * coef [0] [1]
			+ vowel_data_ptr [1] [0]->_freq [fmt] * coef [1] [0]
			+ vowel_data_ptr [1] [1]->_freq [fmt] * coef [1] [1];
		_formants [fmt]._bandwidth =
			  vowel_data_ptr [0] [0]->_bandwidth [fmt] * coef [0] [0]
			+ vowel_data_ptr [0] [1]->_bandwidth [fmt] * coef [0] [1]
			+ vowel_data_ptr [1] [0]->_bandwidth [fmt] * coef [1] [0]
			+ vowel_data_ptr [1] [1]->_bandwidth [fmt] * coef [1] [1];
		_formants [fmt]._gain = _global_param._emphasis;	// We don't care about real formant gain

		_formants [fmt]._filter.set_type (dsp::Filter2p::RESONATOR);
		_formants [fmt]._filter.set_reso (_formants [fmt]._gain);
		_formants [fmt]._filter.set_freq (_formants [fmt]._center_freq);
		_formants [fmt]._filter.set_q (_formants [fmt]._center_freq / _formants [fmt]._bandwidth);
		_formants [fmt]._filter.update_eq ();
	}

	_formants [0]._filter.apply_volume (_filter_gain);

	for (int chn = 0; chn < MAX_POLYPHONY; ++chn)
	{
		_channels [chn].set_update_flag ();
	}

	_need_formant_update_flag = false;
}



void	RectalAnarchy::stop_all_notes ()
{
	for (int track = 0; track < _nbr_tracks; ++track)
	{
		_tracks [track].note_off ();
	}
}



void	RectalAnarchy::stop_all_channels ()
{
	for (int chn = 0; chn < MAX_POLYPHONY; ++chn)
	{
		_channels [chn].deactivate ();
	}
}



const char	RectalAnarchy::COPYRIGHT_0 [] =
	"@(#) Rectal Anarchy - Build " __DATE__ "\n"
	"(c) 2001 FireSledge - Freeware.\n"
	"Laurent de Soras\n"
	"laurent@ohmforce.com -or- ldesoras@club-internet.fr\n"
	"4 avenue Alsace-Lorraine\n"
	"92500 Rueil-Malmaison\n"
	"France";



const char *	RectalAnarchy::VOWEL_NAME_0 [NBR_VOWELS_PER_VOICE] =
{
	"A", "E", "I", "O", "U"
};



const RectalAnarchy::VoiceData	RectalAnarchy::VOICE_LIST [NBR_VOICES] =
{
	// Soprano
	{
		"Soprano",
		{
			{
				{ 800, 1150, 2900, 3900, 4950 },
				{ 0, -6, -32, -20, -50 },
				{ 80, 90, 120, 130, 140 }
			},
			{
				{ 350, 2000, 2800, 3600, 4950 },
				{ 0, -20, -15, -40, -56 },
				{ 60, 100, 120, 150, 200 }
			},
			{
				{ 270, 2140, 2950, 3900, 4950 },
				{ 0, -12, -26, -26, -44 },
				{ 60, 90, 100, 120, 120 }
			},
			{
				{ 450, 800, 2830, 3800, 4950 },
				{ 0, -11, -22, -22, -50 },
				{ 70, 80, 100, 130, 135 }
			},
			{
				{ 325, 700, 2700, 3800, 4950 },
				{ 0, -16, -35, -40, -60 },
				{ 50, 60, 170, 180, 200 }
			}
		}
	},

	// Alto
	{
		"Alto",
		{
			{
				{ 800, 1150, 2800, 3500, 4950 },
				{ 0, -4, -20, -36, -60 },
				{ 80, 90, 120, 130, 140 }
			},
			{
				{ 400, 1600, 2700, 3300, 4950 },
				{ 0, -24, -30, -35, -60 },
				{ 60, 80, 120, 150, 200 }
			},
			{
				{ 350, 1700, 2700, 3700, 4950 },
				{ 0, -20, -30, -36, -60 },
				{ 50, 100, 120, 150, 200 }
			},
			{
				{ 450, 800, 2830, 3500, 4950 },
				{ 0, -9, -16, -28, -55 },
				{ 70, 80, 100, 130, 135 }
			},
			{
				{ 325, 700, 2530, 3500, 4950 },
				{ 0, -12, -30, -40, -64 },
				{ 50, 60, 170, 180, 200 }
			}
		}
	},

	// Countertenor
	{
		"Countertenor",
		{
			{
				{ 660, 1120, 2750, 3000, 3350 },
				{ 0, -6, -23, -24, -38 },
				{ 80, 90, 120, 130, 140 }
			},
			{
				{ 440, 1800, 2700, 3000, 3300 },
				{ 0, -14, -18, -20, -20 },
				{ 70, 80, 100, 120, 120 }
			},
			{
				{ 270, 1850, 2900, 3350, 3590 },
				{ 0, -24, -24, -36, -36 },
				{ 40, 90, 100, 120, 120 }
			},
			{
				{ 430, 820, 2700, 3000, 3300 },
				{ 0, -10, -26, -22, -34 },
				{ 40, 80, 100, 120, 120 }
			},
			{
				{ 370, 630, 2750, 3000, 3400 },
				{ 0, -20, -23, -30, -34 },
				{ 40, 60, 100, 120, 120 }
			}
		}
	},

	// Tenor
	{
		"Tenor",
		{
			{
				{ 650, 1080, 2650, 2900, 3250 },
				{ 0, -6, -7, -8, -22 },
				{ 80, 90, 120, 130, 140 }
			},
			{
				{ 400, 1700, 2600, 3200, 3580 },
				{ 0, -14, -12, -14, -20 },
				{ 70, 80, 100, 120, 120 }
			},
			{
				{ 290, 1870, 2800, 3250, 3540 },
				{ 0, -15, -18, -20, -30 },
				{ 40, 90, 100, 120, 120 }
			},
			{
				{ 400, 800, 2600, 2800, 3000 },
				{ 0, -10, -12, -12, -26 },
				{ 40, 80, 100, 120, 120 }
			},
			{
				{ 350, 600, 2700, 2900, 3300 },
				{ 0, -20, -17, -14, -26 },
				{ 40, 60, 100, 120, 120 }
			}
		}
	},

	// Bass
	{
		"Bass",
		{
			{
				{ 600, 1040, 2250, 2450, 2750 },
				{ 0, -7, -9, -9, -20 },
				{ 60, 70, 110, 120, 130 }
			},
			{
				{ 400, 1620, 2400, 2800, 3100 },
				{ 0, -12, -9, -12, -18 },
				{ 40, 80, 100, 120, 120 }
			},
			{
				{ 250, 1750, 2600, 3050, 3340 },
				{ 0, -30, -16, -22, -28 },
				{ 60, 90, 100, 120, 120 }
			},
			{
				{ 400, 750, 2400, 2600, 2900 },
				{ 0, -11, -21, -20, -40 },
				{ 40, 80, 100, 120, 120 }
			},
			{
				{ 350, 600, 2400, 2675, 2950 },
				{ 0, -20, -32, -28, -36 },
				{ 40, 80, 100, 120, 120 }
			}
		}
	}
};



// Tableau circulaires des valeurs necessaires pour avoir tous les
// enchainements possible entre les nombres de 0 a 4.
const int	RectalAnarchy::TRANSITION_LIST [NBR_TRANSITIONS] =
{
	0, 1, 2, 3, 4, 0, 2, 4, 3, 1
};



const	CMachineParameter	RectalAnarchy::_param_glide_time =
{
	pt_word,								// type
	"VGlide",							// short (one or two words) name 
	"Voice glide time (0 = no glide, 100 = 1 tick, 8000 = 128 ticks)",	// longer description 
	0x0000,								// minumum value
	0xFFFE,								// maximum value
	0xFFFF,								// some value outside [min..max] range 
	MPF_STATE,							// flags
	0x0100								// Default value
};



const	CMachineParameter	RectalAnarchy::_param_voice_type =
{
	pt_word,								// type
	"VType",								// short (one or two words) name 
	"Voice type",						// longer description 
	0x0000,								// minumum value
	0x1000 * NBR_TRANSITIONS - 1,	// maximum value
	0xFFFF,								// some value outside [min..max] range 
	MPF_STATE,							// flags
	0x0000								// Default value
};



const	CMachineParameter	RectalAnarchy::_param_voice_sound =
{
	pt_word,								// type
	"Vowel",								// short (one or two words) name 
	"Vowel",								// longer description 
	0x0000,								// minumum value
	0x1000 * NBR_TRANSITIONS - 1,	// maximum value
	0xFFFF,								// some value outside [min..max] range 
	MPF_STATE,							// flags
	0x0000								// Default value
};



const	CMachineParameter	RectalAnarchy::_param_voice_lfo_depth =
{
	pt_byte,								// type
	"VLFO Dep",							// short (one or two words) name 
	"Voice LFO Depth (0 = off, 80 = 100 %)",	// longer description 
	0x00,									// minumum value
	0x80,									// maximum value
	0xFF,									// some value outside [min..max] range 
	MPF_STATE,							// flags
	0x00									// Default value
};



const	CMachineParameter	RectalAnarchy::_param_voice_lfo_per =
{
	pt_byte,								// type
	"VLFO Per",							// short (one or two words) name 
	"Voice LFO Perdiod (0 = 1/4 tick, 20 = 1 tick, E0 = 4096 ticks)",	// longer description 
	0x00,									// minumum value
	0xE0,									// maximum value
	0xFF,									// some value outside [min..max] range 
	MPF_STATE,							// flags
	0x40									// Default value
};



const	CMachineParameter	RectalAnarchy::_param_voice_lfo_type =
{
	pt_byte,								// type
	"VLFO WF",							// short (one or two words) name 
	"Voice LFO Waveform (0=sin, 1=tri, 2=sq, 3=sawup, 4=sawdwn, 5=bnoise)",	// longer description 
	0x00,									// minumum value
	NBR_LFO - 1,						// maximum value
	0xFF,									// some value outside [min..max] range 
	MPF_STATE,							// flags
	0x00									// Default value
};



const	CMachineParameter	RectalAnarchy::_param_emphasis =
{
	pt_byte,								// type
	"Emphasis",							// short (one or two words) name 
	"Emphasis",							// longer description 
	0x00,									// minumum value
	0xFE,									// maximum value
	0xFF,									// some value outside [min..max] range 
	MPF_STATE,							// flags
	0xC0									// Default value
};



const	CMachineParameter	RectalAnarchy::_param_glottal_open_time =
{
	pt_byte,								// type
	"G.O.Time",							// short (one or two words) name 
	"Glottal Open Time",				// longer description
	0x00,									// minumum value
	0xFE,									// maximum value
	0xFF,									// some value outside [min..max] range 
	MPF_STATE,							// flags
	0x70									// Default value
};



const	CMachineParameter	RectalAnarchy::_param_vca_at =
{
	pt_byte,								// type
	"VCA A",								// short (one or two words) name 
	"VCA Attack time",				// longer description
	0x00,									// minumum value
	0xFE,									// maximum value
	0xFF,									// some value outside [min..max] range 
	MPF_STATE,							// flags
	0x30									// Default value
};



const	CMachineParameter	RectalAnarchy::_param_vca_dt =
{
	pt_byte,								// type
	"VCA D",								// short (one or two words) name 
	"VCA Decay time",					// longer description
	0x00,									// minumum value
	0xFE,									// maximum value
	0xFF,									// some value outside [min..max] range 
	MPF_STATE,							// flags
	0x60									// Default value
};



const	CMachineParameter	RectalAnarchy::_param_vca_sl =
{
	pt_byte,								// type
	"VCA SL",							// short (one or two words) name 
	"VCA Sustain level",				// longer description 
	0x00,									// minumum value
	0x80,									// maximum value
	0xFF,									// some value outside [min..max] range 
	MPF_STATE,							// flags
	0x6C									// Default value
};



const	CMachineParameter	RectalAnarchy::_param_vca_st =
{
	pt_byte,								// type
	"VCA ST",							// short (one or two words) name 
	"VCA Sustain time",				// longer description 
	0x00,									// minumum value
	0xFE,									// maximum value
	0xFF,									// some value outside [min..max] range 
	MPF_STATE,							// flags
	0x00									// Default value
};



const	CMachineParameter	RectalAnarchy::_param_vca_rt =
{
	pt_byte,								// type
	"VCA R",								// short (one or two words) name 
	"VCA Release time",				// longer description 
	0x00,									// minumum value
	0xFE,									// maximum value
	0xFF,									// some value outside [min..max] range 
	MPF_STATE,							// flags
	0x70									// Default value
};



const	CMachineParameter	RectalAnarchy::_param_note =
{
	pt_note,								// type
	"Note",								// short (one or two words) name 
	"Note",								// longer description 
	NOTE_MIN,							// minumum value
	NOTE_MAX,							// maximum value
	NOTE_NO,								// some value outside [min..max] range 
	0,										// flags
	0										// Default value
};



const	CMachineParameter	RectalAnarchy::_param_velo =
{
	pt_byte,								// type
	"Velo",								// short (one or two words) name 
	"Velocity (0 = off, 40 = normal, 7F = max)",	// longer description 
	0x00,									// minumum value
	0x7F,									// maximum value
	0xFF,									// some value outside [min..max] range 
	MPF_STATE,							// flags
	0x40									// Default value
};



const	CMachineParameter	RectalAnarchy::_param_slide =
{
	pt_byte,								// type
	"Slide",								// short (one or two words) name 
	"Slide time (0 = immediate, 4 = 1 tick, 10 = 4 ticks)",	// longer description 
	0x00,									// minumum value
	0xFE,									// maximum value
	0xFF,									// some value outside [min..max] range 
	MPF_STATE,							// flags
	0x00									// Default value
};



const CMachineParameter *	RectalAnarchy::_param_list [NBR_GLOB_PARAM + NBR_TRACK_PARAM] =
{
	// Global param
	&_param_glide_time,
	&_param_voice_type,
	&_param_voice_sound,
	&_param_voice_lfo_depth,
	&_param_voice_lfo_per,
	&_param_voice_lfo_type,
	&_param_emphasis,
	&_param_glottal_open_time,
	&_param_vca_at,
	&_param_vca_dt,
	&_param_vca_sl,
	&_param_vca_st,
	&_param_vca_rt,

	// Track param
	&_param_note,
	&_param_velo,
	&_param_slide
};



const CMachineAttribute	RectalAnarchy::_attrib_sound_quality =
{
	"Sound quality",					// Name
	0,										// Minimum Value
	1,										// Maximum Value
	0										// Default Value
};



const CMachineAttribute	RectalAnarchy::_attrib_polyphony =
{
	"Polyphony",						// Name
	0,										// Minimum Value
	MAX_POLYPHONY,						// Maximum Value
	MAX_POLYPHONY						// Default Value
};



const CMachineAttribute	RectalAnarchy::_attrib_midi_channel =
{
	"MIDI channel",					// Name
	0,										// Minimum Value
	16,									// Maximum Value
	0										// Default Value
};



const CMachineAttribute *	RectalAnarchy::_attrib_list [NBR_ATTRIBUTES] =
{
	&_attrib_sound_quality,
	&_attrib_polyphony,
	&_attrib_midi_channel
};



/*\\\ EOF \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/
