/*****************************************************************************

        RectalAnarchyChannel.cpp
        By Laurent de Soras, 2001

--- Legal stuff ---

This program is free software. It comes without any warranty, to
the extent permitted by applicable law. You can redistribute it
and/or modify it under the terms of the Do What The Fuck You Want
To Public License, Version 2, as published by Sam Hocevar. See
http://sam.zoy.org/wtfpl/COPYING for more details.

*Tab=3***********************************************************************/



#if defined (_MSC_VER)
	#pragma warning (1 : 4130 4223 4705 4706)
	#pragma warning (4 : 4355 4786 4800)
#endif



/*\\\ INCLUDE FILES \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/

#include	"basic/basic_def.h"
#include	"basic/basic_fnc.h"
#include	"RectalAnarchy.h"
#include	"RectalAnarchyChannel.h"
#include	"RectalAnarchyTrack.h"

#include	<cassert>
#include	<cmath>
#include	<climits>
#include	<cstring>



namespace std {};



/*\\\ PUBLIC \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/



RectalAnarchyChannel::RectalAnarchyChannel ()
:	_mch_ptr (0)
,	_track_ptr (0)
,	_sample_freq (44100)
,	_velo (1)
,	_step (440.0f / 44100)
,	_slide_final_step (440.0f / 44100)
,	_glottal_open_time (0.0009f)
,	_update_flag (true)
{
}



RectalAnarchyChannel::~RectalAnarchyChannel ()
{
}



void	RectalAnarchyChannel::init (RectalAnarchy &mch, float sample_freq)
{
	assert (&mch != 0);
	assert (sample_freq > 0);
	assert (_mch_ptr == 0);

	_mch_ptr = &mch;
	_sample_freq = sample_freq;
	_vol_env.set_sample_freq (sample_freq);
	_glotal_lpf.set_sample_freq (sample_freq);
	_glotal_lpf.set_freq (300);
	_glotal_lpf.set_level (0);
	_glotal_lpf.make_high_shelf ();

	for (int fmt = 0; fmt < NBR_FORMANTS; ++fmt)
	{
		_filters [fmt].set_sample_freq (sample_freq);
	}

	set_velo (1);
	set_glottal_open_time (0.0009f);
	_update_flag = true;
}



bool	RectalAnarchyChannel::generate (float data [], long nbr_spl, bool mix_flag)
{
	assert (data != 0);
	assert (nbr_spl > 0);

	const bool		gen_flag = is_active ();

	if (gen_flag)
	{
		using namespace std;

		// Tone slide and misc updates
		if (_step != _slide_final_step)
		{
			_step *= exp (_slide_speed_log * nbr_spl);
			if (   (_slide_speed_log > 0 && _step > _slide_final_step)
			    || (_slide_speed_log < 0 && _step < _slide_final_step))
			{
				_step = _slide_final_step;
			}
		}

		if (_update_flag)
		{
			update_filters ();
		}

		// Generate basic waveform
		float * const	voice_data_ptr = _mch_ptr->pCB->GetAuxBuffer ();
		generate_voice (voice_data_ptr, nbr_spl);

		// Formant filtering
		for (int fmt = 0; fmt < NBR_FORMANTS; ++fmt)
		{
			_filters [fmt].process_block (voice_data_ptr, nbr_spl);
		}

		// Volume envelope
		float * const	vol_data_ptr = _mch_ptr->use_adsr_temp_buf ();
		_vol_env.process_block (vol_data_ptr, nbr_spl);
		long				pos = 0;

		// Mix
		if (mix_flag)
		{
			do
			{
				data [pos] += voice_data_ptr [pos] * vol_data_ptr [pos];
				++pos;
			}
			while (pos < nbr_spl);
		}

		// Replace
		else
		{
			do
			{
				data [pos] = voice_data_ptr [pos] * vol_data_ptr [pos];
				++pos;
			}
			while (pos < nbr_spl);
		}
	}

	return (gen_flag);
}



bool	RectalAnarchyChannel::is_active () const
{
	return (_vol_env.get_cur_section () != dsp::AdsrBase::Section_NONE);
}



void	RectalAnarchyChannel::deactivate ()
{
	_vol_env.clear_buffers ();
	dissociate_from_track ();
}



// [0 ; 1], 0 = free, 1 = can't be deallocated
double	RectalAnarchyChannel::get_activity_rate () const
{
	double			rate;

	if (! is_active ())
	{
		rate = 0;
	}
	else if (_vol_env.get_cur_section () != dsp::AdsrBase::Section_RELEASE)
	{
		rate = 1;
	}
	else
	{
		rate = _vol_env.get_val ();
	}

	return (rate);
}



void	RectalAnarchyChannel::associate_with_track (RectalAnarchyTrack &track)
{
	assert (&track != 0);

	dissociate_from_track ();
	_track_ptr = &track;
}



void	RectalAnarchyChannel::dissociate_from_track ()
{
	if (_track_ptr != 0)
	{
		_track_ptr->dissociate_channel ();
		_track_ptr = 0;
	}
}



// Can be called even if the current state is already "note on" or "note off"
void	RectalAnarchyChannel::note_on (float midi_note, float slide_time)
{
	using namespace std;

	const double	note_freq = 440 * exp ((midi_note - 69) * (basic::LN2 / 12));
	_cur_freq = note_freq;
	_slide_final_step = note_freq / _sample_freq;

	// New note
	if (slide_time < 0)
	{
		_slide_speed_log = 1;
		_vol_env.note_on ();
		_step = _slide_final_step;
		_phase = 0;
		glotal_resync ();
	}

	// Slide
	else
	{
		const float		slide_time_s = slide_time / _mch_ptr->pMasterInfo->TicksPerSec;
		assert (slide_time_s * _sample_freq <= LONG_MAX);
		const long		slide_time_spl = long (slide_time_s * _sample_freq);
		assert (_step != 0);
		if (slide_time_spl == 0)
		{
			_step = _slide_final_step;
			_slide_speed_log = 1;
		}
		else
		{
			const double	mult = _slide_final_step / _step;
			assert (mult > 0);
			_slide_speed_log = log (mult) / slide_time_spl;
		}
	}
}



// Can be called in "note off" state
void	RectalAnarchyChannel::note_off ()
{
	_vol_env.note_off ();
	dissociate_from_track ();
}



void	RectalAnarchyChannel::set_velo (float velo)
{
	assert (velo >= 0);

	_update_flag = true;	// Velocity is part of filtering
	_velo = velo * velo;
}



void	RectalAnarchyChannel::set_glottal_open_time (float t)
{
	assert (t > 0);

	_glottal_open_time = t;
	const float		len_spl = _glottal_open_time * _sample_freq;
	_glotal_len_spl = basic::ceil_int (len_spl);
	_glotal_inc_2 =   (18.0f * RectalAnarchy::BUZZ_SAMPLE_SCALE / 2)
	                / (len_spl * len_spl);
}



void	RectalAnarchyChannel::set_vca_at (float t)
{
	_vol_env.set_attack_time (t);
}



void	RectalAnarchyChannel::set_vca_dt (float t)
{
	_vol_env.set_decay_time (t);
}



void	RectalAnarchyChannel::set_vca_sl (float lvl)
{
	_vol_env.set_sustain_level (lvl);
}



void	RectalAnarchyChannel::set_vca_st (float t)
{
	_vol_env.set_sustain_time (t);
}



void	RectalAnarchyChannel::set_vca_rt (float t)
{
	_vol_env.set_release_time (t);
}



void	RectalAnarchyChannel::set_update_flag ()
{
	_update_flag = true;
}



/*\\\ PROTECTED \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/



/*\\\ PRIVATE \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/



void	RectalAnarchyChannel::update_filters ()
{
	for (int fmt = 0; fmt < NBR_FORMANTS; ++fmt)
	{
		const RectalAnarchy::Formant &	formant = _mch_ptr->get_formant_data (fmt);
		_filters [fmt].copy_filter (formant._filter);
	}
	_filters [0].apply_volume (_velo);
	_update_flag = false;
}



void	RectalAnarchyChannel::generate_voice (float data [], long nbr_spl)
{
	assert (data != 0);
	assert (nbr_spl > 0);

	long				pos = 0;
	do
	{
		const long		spl_before_end = basic::ceil_int ((1 - _phase) / _step);
		assert (spl_before_end > 0);
		long				render_len = basic::min (nbr_spl - pos, spl_before_end);

		if (_glotal_pos > 0)
		{
			render_len = basic::min (render_len, _glotal_pos);
			const long		end_pos = pos + render_len;
			do
			{
				data [pos] = _glotal_lpf.process_sample (_glotal_val);
				_glotal_val += _glotal_inc_1;
				_glotal_inc_1 += _glotal_inc_2;
				++pos;
			}
			while (pos < end_pos);
			_glotal_pos -= render_len;
		}

		else
		{
			const long		end_pos = pos + render_len;
			do
			{
				data [pos] = _glotal_lpf.process_sample (0);
				++pos;
			}
			while (pos < end_pos);
		}

		_phase += _step * render_len;
		while (_phase >= 1)
		{
			_phase -= 1;
			glotal_resync ();
		}
	}
	while (pos < nbr_spl);
}



// Should be called with a valid value set for _phase (0 <= _phase < _step)
void	RectalAnarchyChannel::glotal_resync ()
{
	assert (_phase >= 0);
	assert (_phase < _step);

	const float		phase_spl = float (_phase * _sample_freq / _cur_freq);
	assert (phase_spl >= 0);

	_glotal_inc_1 =   (-6.0f * RectalAnarchy::BUZZ_SAMPLE_SCALE / 2)
	                / (_glottal_open_time * _sample_freq);
	_glotal_val = _glotal_inc_1 * phase_spl;
	_glotal_inc_1 += _glotal_inc_2 * phase_spl;
	_glotal_pos = _glotal_len_spl;
}



/*\\\ EOF \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/
