/*****************************************************************************

        basic_fnc.cpp

--- Legal stuff ---

This program is free software. It comes without any warranty, to
the extent permitted by applicable law. You can redistribute it
and/or modify it under the terms of the Do What The Fuck You Want
To Public License, Version 2, as published by Sam Hocevar. See
http://sam.zoy.org/wtfpl/COPYING for more details.

*Tab=3***********************************************************************/



#if defined (_MSC_VER)
	#pragma warning (1 : 4130 4223 4705 4706)
	#pragma warning (4 : 4355 4786 4800)
#endif



/*\\\ INCLUDE FILES \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/

#include	"basic/basic_fnc.h"

#include <string>

#include	<cassert>
#include	<cctype>
#include	<cstdio>
#include	<cctype>



namespace basic
{



/*\\\ PUBLIC \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/



/*
==============================================================================
Name: gen_string
Description:
	Fill a string with a character.
Input parameters:
	- c: Filling character
	- n: Number of character to fill. >= 0.
Returns: The string.
Throws: ?
==============================================================================
*/

std::string	gen_string (char c, long n)
{
	assert (n >= 0);

	std::string		final_str;
	std::string		filler;
	filler += c;
	while (n != 0)
	{
		if (n & 1)
		{
			final_str += filler;
		}
		filler += filler;
		n /= 2;
	}

	return (final_str);
}



/*
==============================================================================
Name: gen_string
Description:
	Generate a string filled with a pattern.
Input parameters:
	- str_0: Pointer on the pattern string.
	- n: Number of characters in the final string. Thus, if n is not a multiple
		of the str length, the last pattern will be truncated. >= 0
Returns: The patternized string.
Throws: ?
==============================================================================
*/

std::string	gen_string (const char *str_0, long n)
{
	assert (n >= 0);

	using namespace std;

	const long		len = long (strlen (str_0));
	long				n2 = n / len;
	std::string		final_str;
	std::string		filler (str_0);
	while (n2 != 0)
	{
		if ((n2 & 1) != 0)
		{
			final_str += filler;
		}
		filler += filler;
		n2 /= 2;
	}

	final_str += filler.substr (0, n % len);

	return (final_str);
}



/*
==============================================================================
Name: gen_string
Description:
	Generate a string filled with a pattern.
Input parameters:
	- str: The pattern string.
	- n: Number of characters in the final string. Thus, if n is not a multiple
		of the str length, the last pattern will be truncated. >= 0
Returns: The patternized string.
Throws: ?
==============================================================================
*/

std::string	gen_string (const std::string &str, long n)
{
	assert (n >= 0);

	return (gen_string (str.c_str (), n));
}



/*
==============================================================================
Name: print_float
Description:
	Prints a floating point number into a string. with a specified maximum
	length. Numbers after the decimal point are truncated if required, and
	rounded. The decimal delimiter never appears without any number (ex:
	"123." is replaced with "123"). Handles negative numbers.
Input parameters:
	- val: Value to be printed.
	- max_len: Maximum number of character to print, without the trailing null
		character.
Output parameters:
	- dest_0: String with the number, terminated with '\0'.
Throws: Nothing
==============================================================================
*/

void	print_float (char *dest_0, double val, long max_len)
{
	using namespace std;

	assert (dest_0 != 0);
	assert (max_len > 0);

	const double	limit_val = exp (max_len * basic::LN10);
	
#if defined (_DEBUG)
	const double	max_val = limit_val - 0.5;
	const double	min_val = -limit_val * 0.1 + 0.5;

	assert (val > min_val);
	assert (val < max_val);
#endif

	const double	int_max_val = limit_val * 0.01 - 0.5;
	const double	int_min_val = -limit_val * 0.001 + 0.5;

	char				temp_0 [15+1];
	assert (max_len < sizeof (temp_0));

	if (   val >= int_max_val
	    || val <= int_min_val)
	{
		sprintf (temp_0, "%d", basic::round_int (val));
	}
	else
	{
		sprintf (temp_0, "%lf", val);
	}
	assert (strlen (temp_0) < sizeof (temp_0));
	temp_0 [max_len] = '\0';
	strcpy (dest_0, temp_0);
}



/*
==============================================================================
Name: strncpy_0
Description:
	Copies a string and limits the number of characters. The terminating '\0'
	is always there.
Input parameters:
	- src_0: Input string, terminated with '\0'.
	- buf_len: Number of characters, '\0' included.
Output parameters:
	- dest_0: destination location.
Throws: Nothing
==============================================================================
*/

void	strncpy_0 (char *dest_0, const char *src_0, long buf_len)
{
	assert (dest_0 != 0);
	assert (src_0 != 0);
	assert (buf_len >= 0);

	if (buf_len > 0)
	{
		bool				stop_flag = false;
		for (long pos = 0; pos < buf_len && ! stop_flag; ++pos)
		{
			const char		c = src_0 [pos];
			dest_0 [pos] = c;
			stop_flag = (c == '\0');
		}
		dest_0 [buf_len - 1] = '\0';
	}
}



void	strncat_0 (char *dest_0, const char *src_0, long buf_len)
{
	assert (dest_0 != 0);
	assert (src_0 != 0);
	assert (buf_len >= 0);

	if (buf_len > 0)
	{
		using namespace std;

		// Find string end position
		const long		pos = long (strlen (dest_0));
		assert (pos < buf_len);

		// Copy string
		strncpy_0 (&dest_0 [pos], src_0, buf_len - pos);
	}
}



bool	str_jump_to_end (const char * & str_0, long & dec)
{
	char * & str2_0 = const_cast <char * &> (str_0);
	return str_jump_to_end (str2_0, dec);
}		



bool	str_jump_to_end (char * & str_0, long & dec)
{
	using namespace std;
	
	const long		str_len = long (strlen (str_0));
	str_0 += str_len;
	dec -= str_len;
	return (dec <= 0);
}		



/*
==============================================================================
Name: search_str_ci
Description:
	Finds the offset of a string within another (if it exists), case
	insensitive.
Input parameters:
	- s1_0: '\0' terminated string to be found.
	- s2_0: '\0' terminated string to be searched into.
Returns: Offset of s1 in s2. -1 if not found.
Throws: Nothing
==============================================================================
*/

long	search_str_ci (const char *s1_0, const char *s2_0)
{
	using namespace std;

	long				offset = 0;
	for ( ; ; )
	{
		long				cnt = 0;
		while (s1_0 [cnt] != '\0')
		{
			if (s2_0 [cnt + offset] == '\0')
			{
				return (-1);
			}
			if (toupper (s1_0 [cnt]) != toupper (s2_0 [cnt + offset]))
			{
				break;
			}
			++cnt;
		}

		if (s1_0 [cnt] == '\0')
		{
			return (offset);
		}

		++offset;
	}
}



/*
==============================================================================
Name: strcmp_ci
Description:
	Compares two strings for identity, case insensitive.
Input parameters:
	- s1_0: First string
	- s2_0: Second string
Returns: true if equal
Throws: Nothing.
==============================================================================
*/

bool	strcmp_ci (const char *s1_0, const char *s2_0)
{
	using namespace std;

	long				pos = 0;
	bool				ok_flag = true;
	do
	{
		ok_flag = (toupper (s1_0 [pos]) == toupper (s2_0 [pos]));
		++pos;
	}
	while (ok_flag && s1_0 [pos-1] != '\0');

	return (ok_flag);
}



/*
==============================================================================
Name: to_lcase
Description:
	Converts a string to lower case.
Input parameters:
	- txt: String to be converted.
Returns: Converted string in lower case.
Throws: ?
==============================================================================
*/

const std::string	to_lcase (const std::string &txt)
{
	std::string		result (txt);
	to_lcase_inplace (result);

	return (result);
}



/*
==============================================================================
Name: to_ucase
Description:
	Converts a string to upper case.
Input parameters:
	- txt: String to be converted.
Returns: Converted string in upper case.
Throws: ?
==============================================================================
*/

const std::string	to_ucase (const std::string &txt)
{
	std::string		result (txt);
	to_ucase_inplace (result);

	return (result);
}



void	to_lcase_inplace (std::string &txt)
{
	using namespace std;

	const size_t	len = txt.length ();
	for (size_t i = 0; i < len; ++i)
	{
		txt [i] = tolower (txt [i]);
	}
}



void	to_lcase_inplace (char *txt_0)
{
	assert (txt_0 != 0);

	using namespace std;

	const size_t	len = strlen (txt_0);
	for (size_t i = 0; i < len; ++i)
	{
		txt_0 [i] = tolower (txt_0 [i]);
	}
}



void	to_ucase_inplace (std::string &txt)
{
	using namespace std;

	const size_t	len = txt.length ();
	for (size_t i = 0; i < len; ++i)
	{
		txt [i] = toupper (txt [i]);
	}
}



void	to_ucase_inplace (char *txt_0)
{
	assert (txt_0 != 0);

	using namespace std;

	const size_t	len = strlen (txt_0);
	for (size_t i = 0; i < len; ++i)
	{
		txt_0 [i] = toupper (txt_0 [i]);
	}
}



/*
==============================================================================
Name: parse_line
Description:
	Splits a string into tokens, given the provided delimiter characters.
	Substrings between double quotes (") are taken integrally.
Input parameters:
	- line_0: Line to be parsed.
	- delimiter_list_0: Delimiter character list, terminated with '\0'.
Input/output parameters:
	- lex_arr: Token list. If it is not empty when the function is called, the
		found token are pushed back to the list.
Returns: Number of produced tokens.
Throws: ?
==============================================================================
*/

int	parse_line (std::vector <std::string> &lex_arr, const char *line_0, const char *delimiter_list_0)
{
	assert (line_0 != 0);
	assert (delimiter_list_0 != 0);

	int				pos = 0;
	int				lex_pos = 0;
	int				lex_cnt = 0;
	bool				quote_flag = false;
	bool				argument_flag = false;
	std::string		lexem;

	while (line_0 [pos] != '\0')
	{
		const char		c = line_0 [pos];
		const bool		delimiter_flag = is_delimiter (c, delimiter_list_0);

		if (quote_flag)
		{
			if (! argument_flag)
			{
				lex_pos = pos;
				argument_flag = true;
			}

			if (c == '\"')
			{
				lex_arr.push_back (std::string (line_0 + lex_pos, pos - lex_pos));
				++ lex_cnt;
				quote_flag = false;
				argument_flag = false;
			}
		}

		else
		{
			if (c == '\"')
			{
				if (argument_flag)
				{
					lex_arr.push_back (std::string (line_0 + lex_pos, pos - lex_pos));
					++ lex_cnt;
					argument_flag = false;
				}

				quote_flag = true;
			}

			else if (argument_flag)
			{
				if (delimiter_flag)
				{
					lex_arr.push_back (std::string (line_0 + lex_pos, pos - lex_pos));
					++ lex_cnt;
					argument_flag = false;
				}
			}

			else if (! delimiter_flag)
			{
				lex_pos = pos;
				argument_flag = true;
			}
		}

		++ pos;
	}

	if (argument_flag)
	{
		lex_arr.push_back (std::string (line_0 + lex_pos, pos - lex_pos));
		++ lex_cnt;
	}

	return (lex_cnt);
}



/*
==============================================================================
Name: is_delimiter
Description:
	Check if a character is in a (delimiter) list.
Input parameters:
	- c: Character to check
	- delimiter_list_0: Delimiter character list, terminated with '\0'.
Returns: true if the character belongs to the list.
Throws: Nothing
==============================================================================
*/

bool	is_delimiter (char c, const char *delimiter_list_0)
{
	while (*delimiter_list_0 != '\0')
	{
		if (c == *delimiter_list_0)
		{
			return (true);
		}
		++delimiter_list_0;
	}

	return (false);
}



/*
==============================================================================
Name: test_filename
Description:
	Tests if a filename corresponds to a supplied mask. If the name has no
	extension and no dot, the search will be validated only there is a '*'
	jocker in the mask.
Input parameters:
	- filename_0: Filename, terminated with '\0'.
	- mask_0: Mask string, terminated with '\0'. Can contain jockers like
		'?' (any character) or '*' (any substring). Multiple mask can be
		tested, they must be separated with ',' (thus ',' cannot be a match
		character for a mask). A mask should not contain any extra space.
Returns: true if the name matches the mask.
Throws: Nothing
==============================================================================
*/

bool	test_filename (const char *filename_0, const char *mask_0)
{
	using namespace std;

	int				char_nbr;
	int				length;
	char *			temp_0;
	char *			virgule_0;

	// -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -
	// Multiple masks

	// With VC8, there is a symbol resolution problem if we omit the (char*) cast
	while ((virgule_0 = (char*)strchr (mask_0, ',')) != 0)
	{
		length = int (virgule_0 - mask_0);
		std::string		mask_2 (mask_0, length);
		if (test_filename (filename_0, mask_2.c_str ()))
		{
			return (true);
		}
		mask_0 = virgule_0 + 1;
	}

	// -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -
	// Search on a simple mask

	while (*mask_0 != 0)
	{
		if (*filename_0 == 0)
		{
			if (*mask_0++ == '.')
			{
				while (*mask_0 == '*')
				{
					mask_0 ++;
				}
				if (*mask_0 == 0)
				{
					return (true);
				}
			}
			return (false);
		}

		if (*mask_0 == '\?')
		{
			filename_0 ++;
		}

		else if (*mask_0 == '*')
		{
			while (mask_0 [1] == '*')
			{
				mask_0 ++;
			}
			char_nbr = 0;
			while (mask_0 [1] == '\?')
			{
				char_nbr ++;
				mask_0 ++;
			}
			if (mask_0 [1] == 0)
			{
				return (true);
			}

			// VC8 (char*) cast
			temp_0 = (char *) strchr (filename_0, toupper (mask_0 [1]));
			if (temp_0 == 0)
			{
				// VC8 (char*) cast
				temp_0 = (char *) strchr (filename_0, tolower (mask_0 [1]));
			}

			if (temp_0 == 0)
			{
				if (   mask_0 [1] == '.'
				    && mask_0 [2] == '*'
				    && strlen (mask_0) == 3)
				{
					return (true);
				}

				return (false);
			}
			if (temp_0 - filename_0 < char_nbr)
			{
				return (false);
			}
			filename_0 = temp_0;
		}

		else
		{
			if (toupper (*filename_0) != toupper (*mask_0))
			{
				if (   mask_0 [1] == '.'
				    && mask_0 [2] == '*'
				    && strlen (mask_0) == 3)
				{
					return (true);
				}

				return (false);
			}
			filename_0 ++;
		}

		mask_0 ++;
	}

	if (*filename_0 != 0)
	{
		return (false);
	}
		
	return (true);
}



void	trim_string (std::string &txt, const char *delimiter_list_0)
{
	assert (&txt != 0);
	assert (delimiter_list_0 != 0);

	while (   txt.length () > 0
	       && is_delimiter (txt [txt.length () - 1], delimiter_list_0))
	{
		txt = txt.substr (0, txt.length () - 1);
	}

	while (   txt.length () > 0
	       && is_delimiter (txt [0], delimiter_list_0))
	{
		txt = txt.substr (1);
	}
}



/*
==============================================================================
Name: compact_string_for_display
	Compact a string to fit in a given size. This is useful for example for
	pathnames, when one need to see both beginning and end. The string is
	truncated by removing the middle part and replacing it by "...".
	Delimiters help to isolate words and keep their integrity, if possible.
Description:
Input parameters:
	- txt: String to compact
	- max_len: Maximum output string length. >= 0
	- delimiter_list_0: list of delimiter between "tokens". May be an empty
		string.
Returns: Truncated string. Its length is less or equal to max_len.
Throws: std::string related exceptions
==============================================================================
*/

std::string	compact_string_for_display (const std::string &txt, long max_len, const char *delimiter_list_0)
{
	assert (&txt != 0);
	assert (max_len >= 0);
	assert (delimiter_list_0 != 0);

	const long		dot_len = 3;

	std::string		shrink;
	const long		txt_len = long (txt.length ());
	if (txt_len <= max_len)
	{
		shrink = txt;
	}

	else if (max_len <= 1 + dot_len + 1)
	{
		shrink = txt.substr (0, max_len);
	}

	else
	{
		typedef	std::vector <long>	PosList;

		// Find the delimiter positions
		PosList			delim_pos_list;
		for (long pos = 0; pos < txt_len; ++pos)
		{
			if (is_delimiter (txt [pos], delimiter_list_0))
			{
				delim_pos_list.push_back (pos);
			}
		}
		const long		nbr_delim = long (delim_pos_list.size ());

		// Try to get a pair of acceptable delimiters
		bool				brutal_trunc_flag = (nbr_delim < 2);
		if (! brutal_trunc_flag)
		{
			int				left = (nbr_delim - 1) / 2;
			int				right = left + 1;
			assert (left >= 0);
			assert (right < nbr_delim);

			bool				found_flag = false;
			bool				modif_left_side_flag = true;
			while (left >= 0 && right < nbr_delim && ! found_flag)
			{
				const long		left_pos = delim_pos_list [left];
				const long		left_len = left_pos + 1;	// Includes delim
				const long		right_pos = delim_pos_list [right];
				const long		right_len = txt_len - right_pos;	// Includes delim
				const long		shrink_len = left_len + dot_len + right_len;
				if (shrink_len <= max_len)
				{
					found_flag = true;
				}
				else
				{
					if (modif_left_side_flag)
					{
						-- left;
					}
					else
					{
						++ right;
					}
					modif_left_side_flag = ! modif_left_side_flag;
				}
			}

			if (! found_flag)
			{
				brutal_trunc_flag = true;
			}
			else
			{
				const long		left_len = delim_pos_list [left] + 1;	// Includes delim
				const long		right_pos = delim_pos_list [right];
				const long		right_len = txt_len - right_pos;	// Includes delim

				shrink =   txt.substr (0, left_len)
				         + gen_string ('.', dot_len)
				         + txt.substr (right_pos, right_len);
			}
		}

		if (brutal_trunc_flag)
		{
			const long		left_len = (max_len - dot_len) / 2;
			const long		right_len = max_len - (left_len + dot_len);
			const long		right_pos = txt_len - right_len;

			shrink =   txt.substr (0, left_len)
			         + gen_string ('.', dot_len)
			         + txt.substr (right_pos, right_len);
		}
	}

	assert (static_cast <long> (shrink.length ()) <= max_len);

	return (shrink);
}



/*
==============================================================================
Name: decode_compile_date
Description:
	Turns a __DATE__ into somthing usable.
Input parameters:
	- date_0: The __DATE__ string (from the compiler macro).
Output parameters:
	- year: Year, in [1950 ; 2999[
	- month: [0 ; 11]
	- day: [0 ; 27] to [0 ; 31] depending on the month.
Throws: Nothing
==============================================================================
*/

void	decode_compile_date (int &year, int &month, int &day, const char *date_0)
{
	assert (date_0 != 0);

	using namespace std;

	const int		nbr_months = 12;
	static const char *	month_list_0 [nbr_months] =
	{
		"Jan", "Feb", "Mar", "Apr", "May", "Jun",
		"Jul", "Aug", "Sep", "Oct", "Nov", "Dec"
	};
	assert (month_list_0 [nbr_months - 1] != 0);

	char				copy_0 [15+1];
	strncpy_0 (copy_0, date_0, 3+1);
	month = -1;
	for (int test_m = 0; test_m < nbr_months && month < 0; ++test_m)
	{
		if (basic::search_str_ci (month_list_0 [test_m], date_0) == 0)
		{
			month = test_m;
		}
	}

	char *			year_0;
	day = int (strtol (date_0 + 4, &year_0, 10)) - 1;
	year = atoi (year_0);

	assert (day >= 0);
	assert (day < 31);
	assert (month >= 0);
	assert (month < 12);
	assert (year >= 1950);	// Say this limit is OK.
	assert (year < 3000);
}



/*\\\ PROTECTED \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/



/*\\\ PRIVATE \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/



/*

(Legacy notes, useless now with SSE & co)

Optimisations a creuser, sans branchement.
Les erreurs d'arrondi ont deux consequences :
- Introduction d'erreurs, particulierement pour les valeurs proches de 0
- Elimination de la denormalisation si les bornes sont differentes de 0 et
que les resultats intermediaires des calculs sont stockes avec une precision
limitee (elimination a -400 dB pour du 80 bits, -144 pour du 32 bits).

max (x, a) :

	x -= a;
	x += fabs (x);
	x *= 0.5;
	x += a;

min (x, b) :

	x = b - x;
	x += fabs (x)
	x *= 0.5;
	x = b - x;

limit (x, a, b) :

	x -= a;
	x += fabs (x);
	x = (b - a)*2 - x;
	x += fabs (x);
	x *= 0.25;
	x = b - x;

limit (x, a, b) (plus d'operations en parallele) :

	x1 = fabs (x-a);
	x2 = fabs (x-b);
	x = x1 + (a+b)
	x -= x2
	x *= 0.5;
*/



}	// namespace basic



/*\\\ EOF \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/
