/*****************************************************************************

        basic_fnc.hpp

--- Legal stuff ---

This program is free software. It comes without any warranty, to
the extent permitted by applicable law. You can redistribute it
and/or modify it under the terms of the Do What The Fuck You Want
To Public License, Version 2, as published by Sam Hocevar. See
http://sam.zoy.org/wtfpl/COPYING for more details.

*Tab=3***********************************************************************/



#if ! defined (basic_basic_fnc_CODEHEADER_INCLUDED)
#define	basic_basic_fnc_CODEHEADER_INCLUDED



/*\\\ INCLUDE FILES \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/

#include	"archi/archi_def.h"
#include	"basic/basic_def.h"

#include	<cassert>
#include	<cmath>
#include	<cstring>
#include	<cctype>

#if defined (_MSC_VER)

#include	<intrin.h>

#if (archi_WORD_SIZE == 64)
	#pragma intrinsic (_BitScanReverse)
#endif

#elif defined (__POWERPC__) && defined (__GNUC__)

#include <ppc_intrinsics.h>

#endif



namespace std {}



namespace basic
{



/*\\\ PUBLIC \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/



inline float	map_lin_2_exp (float val, float k);



/*
==============================================================================
Name: max
Description:
	Returns the maximum of two value (compared with operator < )
Input parameters:
	- a: First value
	- b: Second value
Returns: The greater value
Throws: Depends on the type (nothing for PODs)
==============================================================================
*/

template <class T>
T	max (T a, T b)
{
	return (a < b ? b : a);
}



template <class T>
void	max_in_place (T &a, T b)
{
	if (a < b)
	{
		a = b;
	}
}



/*
==============================================================================
Name: min
Description:
	Returns the minimum of two value (compared with operator < )
Input parameters:
	- a: First value
	- b: Second value
Returns: The lower value.
Throws: Depends on the type (nothing for PODs)
==============================================================================
*/

template <class T>
T	min (T a, T b)
{
	return (b < a ? b : a);
}



template <class T>
void	min_in_place (T &a, T b)
{
	if (b < a)
	{
		a = b;
	}
}



template <class T>
T	abs (T a)
{
	return (a < 0 ? -a : a);
}

template <>
float	abs (float a)
{
	using namespace std;
	
	return (static_cast <float> (fabs (a)));
}

template <>
double	abs (double a)
{
	using namespace std;
	
	return (fabs (a));
}



/*
==============================================================================
Name: limit
Description:
	Clips a value between two bounds.
Input parameters:
	- a: Value to be clipped
	- b: Lower bound
	- c: Upper bound
Returns: The clipped value.
Throws: Depends on the type (nothing for PODs)
==============================================================================
*/

template <class T>
T	limit (T a, T b, T c)
{
	assert (! (c < b));

	return (max (min (a, c), b));
}



/*
==============================================================================
Name: limit_in_place
Description:
	Clips a value between two bounds.
Input parameters:
	- b: Lower bound
	- c: Upper bound
Input/output parameters:
	- a: Value to be clipped
Throws: Depends on the type (nothing for PODs)
==============================================================================
*/

template <class T>
void	limit_in_place (T &a, T b, T c)
{
	assert (! (c < b));

	if (a < b)
	{
		a = b;
	}
	else if (c < a)
	{
		a = c;
	}
}



/*
==============================================================================
Name: is_null
Description:
	Indicates if a floating point value can be considered as 0. Actually,
	its absolute value is compared to EPSILON.
Input parameters:
	- x: Value whose nullity is to be known.
Returns: true if x can be considered as 0.
Throws: Nothing
==============================================================================
*/

bool	is_null (double x)
{
	using namespace std;

	return (fabs (x) < basic::EPSILON);
}



/*
==============================================================================
Name: is_equal
Description:
	Fuzzy equality test.
Input parameters:
	- x: First number to compare
	- y: Second number to compare
Returns: true if the values can be considered as equal.
Throws: Nothing
==============================================================================
*/

bool	is_equal (double x, double y)
{
	return (is_equal_tol (x, y, basic::EPSILON));
}



bool	is_equal_tol (double x, double y, double d)
{
	using namespace std;
	assert (d >= 0);

	return (is_equal_delta (x, y, fabs (x) * d));
}



bool	is_equal_delta (double x, double y, double d)
{
	assert (d >= 0);

	using namespace std;

	return (fabs (x - y) <= d);
}



bool	is_denormal (const float &f)
{
	assert (sizeof (int) == sizeof (float));
	const int		x = *reinterpret_cast <const int *> (&f);

	return (   (x & 0x7f800000) == 0
	        && (x & 0x807FFFFF) != 0);
}



/*
==============================================================================
Name: round
Description:
	Rounds a number to the closest integer, rounding 0.5 to the upper integer.
Input parameters:
	- x: Value to be rounded.
Returns: The rounded value
Throws: Nothing
==============================================================================
*/

double	round (double x)
{
	using namespace std;

	return (floor (x + 0.5));
}



/*
==============================================================================
Name: round_int
Description:
	Rounds a number to the closest integer, rounding 0.5 to the upper integer.
	Then, the value is converted to integer.
Input parameters:
	- x: Value to be rounded. See assert() for limits.
Returns: The rounded value
Throws: Nothing
==============================================================================
*/

int	round_int (double x)
{
	assert (x <= double (INT_MAX));
	assert (x >= double (INT_MIN));

#if defined (_MSC_VER)

	assert (x <= double (INT_MAX/2));
	assert (x >= double (INT_MIN/2));

	int				i;
	static const float	round_to_nearest = 0.5f;

#if archi_WORD_SIZE == 64

	const double	xx = x + x + round_to_nearest;
	const __m128d	x_128d = _mm_set_sd (xx);
	i = _mm_cvtsd_si32 (x_128d);
	i >>= 1;

#else

	__asm
	{
		fld				x
		fadd				st, st (0)
		fadd				round_to_nearest
		fistp				i
		sar				i, 1
	}

#endif

	assert (i == int (round (x)));

	return (i);

#elif defined(__GNUC__) && defined (__i386__) 

	assert (x <= double (INT_MAX/2));
	assert (x >= double (INT_MIN/2));

	int				i;

	static const float	round_to_nearest = 0.5f;
	asm (
		"fldl				%[x]				\n"
		"fadd				%%st (0), %%st	\n"
		"fadd				%[r]				\n"
		"fistpl			%[i]				\n"
		"sarl				$1, %[i]			\n"
	:	[i]	"=m"	(i)
	:	[r]	"m"	(round_to_nearest)
	,	[x]	"m"	(x)
	:	"st"
	);

	using namespace std;
	assert (i == int (round (x)));

	return (i);

#else

	return (int (round (x)));

#endif
}



// Doesn't give the right result for very small negative values.
int	floor_int (double x)
{
	assert (x <= double (INT_MAX));
	assert (x >= double (INT_MIN));

#if defined (_MSC_VER)

	assert (x <= double (INT_MAX/2));
	assert (x >= double (INT_MIN/2));

	int				i;
	static const float	round_toward_m_i = -0.5f;

#if (archi_WORD_SIZE == 64)

	const double	xx = x + x + round_toward_m_i;
	const __m128d	x_128d = _mm_set_sd (xx);
	i = _mm_cvtsd_si32 (x_128d);
	i >>= 1;

#else

	__asm
	{
		fld				x
		fadd				st, st (0)
		fadd				round_toward_m_i
		fistp				i
		sar				i, 1
	}

#endif

	using namespace std;
	assert (i == int (floor (x)) || fabs (i - x) < 1e-10);

	return (i);

#elif defined(__GNUC__) && defined(__i386__) 

	assert (x <= double (INT_MAX/2));
	assert (x >= double (INT_MIN/2));

	int				i;
//	int *				i_ptr = &i;
	
	static const float	round_toward_m_i = -0.5f;
	asm (
		"fldl				%[x]				\n"
		"fadd				%%st (0), %%st	\n"
		"fadd				%[rm]				\n"
		"fistpl			%[i]				\n"
		"sarl				$1, %[i]			\n"
	:	[i]	"=m"	(i)
	:	[rm]	"m"	(round_toward_m_i)
	,	[x]	"m"	(x)
	:	//"st"
	);

	using namespace std;
	assert (i == int (floor (x)) || fabs (i - x) < 1e-10);

	return (i);

#else

	using namespace std;
	return (int (floor (x)));

#endif
}



// Doesn't give the right result for very small positive values.
int	ceil_int (double x)
{
	assert (x <= double (INT_MAX));
	assert (x >= double (INT_MIN));

#if defined (_MSC_VER)

	assert (x <= double (INT_MAX/2));
	assert (x >= double (INT_MIN/2));

	int				i;
	static const float	round_toward_p_i = -0.5f;

#if archi_WORD_SIZE == 64

	const double	xx = round_toward_p_i - (x + x);
	const __m128d	x_128d = _mm_set_sd (xx);
	i = _mm_cvtsd_si32 (x_128d);
	i >>= 1;

#else

	__asm
	{
		fld				x
		fadd				st, st (0)
		fsubr				round_toward_p_i
		fistp				i
		sar				i, 1
	}

#endif

	using namespace std;
	assert (-i == int (ceil (x)) || fabs (-i - x) < 1e-10);

	return (-i);

#elif defined(__GNUC__) && defined(__i386__) 

	assert (x <= double (INT_MAX/2));
	assert (x >= double (INT_MIN/2));

	int				i;

	static const float	round_toward_p_i = -0.5f;
	asm (
		"fldl				%[x]				\n"
		"fadd				%%st (0), %%st	\n"
		"fsubr			%[r]				\n"
		"fistpl			%[i]				\n"
		"sarl				$1, %[i]			\n"
	:	[i]	"=m"	(i)
	:	[r]	"m"	(round_toward_p_i)
	,	[x]	"m"	(x)
	:	//"st"
	);

	using namespace std;
	assert (-i == int (ceil (x)) || fabs (-i - x) < 1e-10);

	return (-i);

#else

	using namespace std;
	return (int (ceil (x)));

#endif
}



int	float_to_int_fast (float x)
{
	assert (x <= INT_MAX);
	assert (x >= INT_MIN);

	int				val;

#if defined (_MSC_VER)

	#if (archi_WORD_SIZE == 64)

	val = float_to_int_fast (double (x));

	#else

	__asm
	{
		fld				x;
		fistp				val;
	}

	#endif

#elif defined(__GNUC__) && defined(__i386__) 

	asm (
		"flds				%[x]		\n"
		"fistpl			%[v]		\n"
	:	[v]	"=m"	(val)
	:	[x]	"m"	(x)
	:	
	);

#else

	val = int (x);

#endif

	return (val);
}



int	float_to_int_fast (double x)
{
	assert (x <= INT_MAX);
	assert (x >= INT_MIN);

	int				val;

#if defined (_MSC_VER)

	#if (archi_WORD_SIZE == 64)

	const __m128d	x_128d = _mm_set_sd (x);
	val = _mm_cvtsd_si32 (x_128d);

	#else

	__asm
	{
		fld				x;
		fistp				val;
	}

	#endif

#elif defined (__GNUC__) && defined(__i386__) 

	asm (
		"fldl				%[x]		\n"
		"fistpl			%[v]		\n"
	:	[v]	"=m"	(val)
	:	[x]	"m"	(x)
	:	
	);

#else

	val = int (x);

#endif

	return (val);
}



// Faster than round_int() but a bit less accurate (for the x.5 things)
int	round_int_fast (float x)
{
	assert (x <= INT_MAX);
	assert (x >= INT_MIN);

	int				val;

#if defined (_MSC_VER)

	#if (archi_WORD_SIZE == 64)

	const __m128d	x_128d = _mm_set_sd (x);
	val = _mm_cvtsd_si32 (x_128d);

	#else

	__asm
	{
		fld				x;
		fistp				val;
	}

	#endif

	return (val);

#elif defined (__GNUC__) && defined (__i386__) 

	asm (
		"flds				%[x]		\n"
		"fistpl			%[v]		\n"
	:	[v]	"=m"	(val)
	:	[x]	"m"	(x)
	:	
	);

#else

	val = round_int (x);

#endif

	return (val);
}



template <class T>
int	sgn (T x)
{
	return (  (0 < x)
	        ? 1
	        : (  (x < 0)
	           ? -1
				  : 0));
}



double	sinc (double x)
{
	if (x == 0)
	{
		return (1);
	}

	using namespace std;

	const double	pi_x = x * basic::PI;

	return (sin (pi_x) / pi_x);
}



double	acosh (double x)
{
	assert (x >= 1);

	using namespace std;

	const double	tmp = x + sqrt (x*x - 1);
	assert (tmp > 0);

	return (log (tmp));
}



double	asinh (double x)
{
	using namespace std;

	const double	tmp = x + sqrt (x*x + 1);
	assert (tmp > 0);

	return (log (tmp));
}



double	atanh (double x)
{
	assert (fabs (x) < 1);

	const double	tmp = (1 + x) / (1 - x);
	assert (tmp > 0);

	using namespace std;

	return (0.5 * log (tmp));
}



template <class T>
T	ipow (T x, long n)
{
	unsigned long	abs_n = basic::abs (n);
	const T			z (ipowp (x, abs_n));

	return ((n < 0) ? T (1) / z : z);
}



template <class T>
T	ipowp (T x, long n)
{
	assert (n >= 0);

	T					z (1);
	while (n != 0)
	{
		if ((n & 1) != 0)
		{
			z *= x;
		}
		n >>= 1;
		x *= x;
	}

	return (z);
}



float	fast_log2 (float val)
{
	assert (val > 0);

	int * const		exp_ptr = reinterpret_cast <int *> (&val);
	int				x = *exp_ptr;
	const int		log_2 = ((x >> 23) & 255) - 128;
	x &= ~(255 << 23);
	x += 127 << 23;
	*exp_ptr = x;

	val = map_lin_2_exp (val, 1.0f / 2);	// Optionnal

	return (val + log_2);
}



float	fast_exp2 (float val)
{
  // Store address of float as long pointer
  long *				p = (long*) (&val);

  // Truncated val for integer power of 2
  const long		tx = floor_int (val);

  // Float remainder of power of 2
  val -= static_cast <float> (tx);

  // Quadratic approximation of 2^x in [0 ; 1]
  const float		a = 1.0f / 3.0f;
  const float		b = 2.0f / 3.0f;
  const float		c = 1.0f;
  val = (a * val + b) * val + c;

  // Add integer power of 2 to exponent
  *p += (tx << 23);

  assert (val >= 0);

  return (val);
}



// (1,1)-(2,2) => (1,1)-(2,2)
// Beware the bounds, there could be calculation noise errors.
float	map_lin_2_exp (float val, float k)
{
//	assert (val >= 1);
//	assert (val <= 2);
	assert (val >= 0.9);		// Because of calculation noise
	assert (val <= 2.1);

	const float		a = (k - 1) / (k + 1);
	const float		b = (4 - 2*k) / (k + 1);	// 1 - 3*a
	const float		c = 2*a;
	val = (a * val + b) * val + c;

//	assert (val >= 1);
//	assert (val <= 2);
	assert (val >= 0.9);		// Because of calculation noise
	assert (val <= 2.1);

	return (val);
}



/*
==============================================================================
Name : fast_partial_exp2_int_16_to_int_32
Description :
	Compute a fixex-point approximation of the following function :
	[0 ; 1[ -> [0.5 ; 1[
	  f : x -> 2^(x-1)
	The approximation uses the following polynomial :
	f(x) = (((x+1)^2) + 2) / 6
Input parameters :
	- val : Unsigned linear input, Fixed-point 0:16 data. Integer part
		is automatically discarded.
Returns: Unsigned exponential output, fixed-point 0:32 data, range is
	[ 0x80000000 ; 0xFFFFFFFF ]
Throws : Nothing
==============================================================================
*/

inline archi::UInt32	fast_partial_exp2_int_16_to_int_32 (int val)
{
	archi::UInt32	result;

#if defined (_MSC_VER) && archi_WORD_SIZE == 32

	__asm
	{
		mov				eax, val
		shl				eax, 15				; eax = pitch_frac [31/31 bits]
		or					eax, 080000000h	; eax = pitch_frac + 1  [32/31 bits]
		mul				eax
		mov				eax, edx				; eax = (pitch_frac + 1) ^ 2 [32/30 bits]
		mov				edx, 2863311531	; 2/3 [32 bits], rounded to +oo
		mul				edx					; eax = 2/3 (pitch_frac + 1) ^ 2 [32/30 bits]
		add				edx, 1431655766	; + 4/3 [30 bits] + 1
		mov				result, edx
	}

#elif defined (archi_SYS_BE) || (defined (__i386__) && defined (__GNUC__))

	__asm__ ("	 											\n	\
		shll				$15, 				%0				\n	\
		orl				$0x80000000,	%0				\n	\
		movl				%0,				%%eax			\n	\
		mull				%%eax								\n	\
		movl				%%edx,			%%eax			\n	\
		movl				$2863311531,	%%edx			\n	\
		mull				%%edx								\n	\
		addl				$1431655766,	%%edx			\n	\
		movl				%%edx,			%1 			\n	\
		"
		: "=r"(result)
		: "0" (val)
		:	"%eax", "%edx"
	);
	
#elif defined (__POWERPC__) && (defined (__MWERKS__) || defined (__GNUC__))
	
	result = val;
	result <<= 15;
	result |= 0x80000000;
	result = __mulhwu (result, result);
	result = __mulhwu (result, 2863311531UL);
	result += 1431655766UL;

#else

	const int		resol = 16;
	const archi::UInt32	mask = (1 << resol) - 1;

	val &= mask;
	archi::Int64	step_64 (val + (1 << resol));
	step_64 *= step_64;
	step_64 *= (1 << (62 - (resol + 1) * 2)) / 3 + 1820;
	result = (archi::UInt32) (step_64 >> (62 - 32 - 1));
	result += 1431641208;
	
#endif
	
	assert ((result & 0x80000000UL) != 0);

	return (result);
}



// Way more accurate (error < 8e-5 -> ~0.08 cent)
// Polynomial approximation by Olli Niemitalo
// y = (((1/150 * x + 2/75) * x + 3/25) * x + 26/75) * x + 0.5
archi::UInt32	fast_partial_exp2_int_16_to_int_32_4th (int val)
{
	archi::UInt32	result;

#if defined (WIN32) && archi_WORD_SIZE == 32

	__asm
	{
		mov				ecx, val
		shl				ecx, 16
		mov				edx, 001B4E81Bh	; 1/150
		mov				eax, ecx
		mul				edx					; edx = 1/150 * x
		add				edx, 006D3A06Dh	; 2/75
		mov				eax, edx
		mul				ecx					; (1/150 * x + 2/75) * x
		add				edx, 01EB851EBh	; 3/25
		mov				eax, edx
		mul				ecx					; (1/150 * x + 2/75) * x + 3/25) * x
		add				edx, 058BF258Ch	; 26/75
		mov				eax, edx
		mul				ecx					; (((1/150 * x + 2/75) * x + 3/25) * x + 26/75) * x
		add				edx, 080000000h	; (((1/150 * x + 2/75) * x + 3/25) * x + 26/75) * x + 0.5
		mov				result, edx
	}

#elif defined (WIN32) && defined (__GNUC__)

	__asm__ ("												\n	\
		shll				$16, 				%0				\n \
		movl				$0x01B4E81B,	%%edx			\n	\
		movl				%0,				%%eax			\n	\
		mull				%%edx								\n	\
		addl				$0x06D3A06D,	%%edx 		\n	\
		movl				%%edx,			%%eax			\n	\
		mull				%0									\n	\
		addl				$0x1EB851EB,	%%edx 		\n	\
		movl				%%edx,			%%eax			\n	\
		mull				%0									\n	\
		addl				$0x58BF258C,	%%edx 		\n	\
		movl				%%edx,			%%eax			\n	\
		mull				%0									\n	\
		addl				$0x80000000,	%%edx 		\n	\
		movl				%%edx,			%1 			\n	\
		"
		: "=r"(result)
		: "0" (val)
		:	"%eax", "%edx"
	);

			
#elif defined (__POWERPC__) && (defined (__MWERKS__) || defined (__GNUC__))
	
	archi::UInt32	x;

	x = val;
	x <<= 16;
	result = __mulhwu (x, 0x01B4E81B);
	result += 0x06D3A06D;
	result = __mulhwu (result, x);
	result += 0x1EB851EB;
	result = __mulhwu (result, x);
	result += 0x58BF258C;
	result = __mulhwu (result, x);
	result += 0x80000000;

#elif defined (archi_SYS_BE) || (defined (__i386__) && defined (__GNUC__))

	__asm__ ("	 											\n	\
		shll				$16, 				%0				\n	\
		movl				$0x01B4E81B,	%%edx			\n	\
		movl				%0,				%%eax			\n	\
		mull				%%edx								\n	\
		addl				$0x06D3A06D,	%%edx 		\n	\
		movl				%%edx,			%%eax 		\n	\
		mull				%0									\n	\
		addl				$0x1EB851EB,	%%edx 		\n	\
		movl				%%edx,			%%eax			\n	\
		mull				%0									\n	\
		addl				$0x58BF258C,	%%edx 		\n	\
		movl				%%edx,			%%eax 		\n	\
		mull				%0									\n	\
		addl				$0x80000000,	%%edx 		\n	\
		movl				%%edx				,%1 			\n	\
		"
		: "=r"(result)
		: "0" (val)
		:	"%eax", "%edx"
	);

#else

	const archi::UInt32	x_32 = static_cast <archi::UInt32> (val) << 16;
	result = static_cast <archi::UInt32> ((archi::UInt64 (x_32) * 0x001B4E81B) >> 32);
	result += 0x06D3A06D;
	result = static_cast <archi::UInt32> ((result * archi::UInt64 (x_32)) >> 32);
	result += 0x1EB851EB;
	result = static_cast <archi::UInt32> ((result * archi::UInt64 (x_32)) >> 32);
	result += 0x58BF258C;
	result = static_cast <archi::UInt32> ((result * archi::UInt64 (x_32)) >> 32);
	result += 0x80000000;

#endif

	assert ((result & 0x80000000UL) != 0);

	return (result);
}



float	tan_taylor_5th (float val)
{
	const float		val_2 = val * val;
	const float		c_1 = 1;
	const float		c_3 = 1.0f / 3;
	const float		c_5 = 2.0f / 15;

	return (((c_5 * val_2 + c_3) * val_2 + c_1) * val);
}



/*
==============================================================================
Name: pseudo_exp
Description:
	Computes a pseudo-exponent of a number x. It gives 0 for x = 0, and 1 for
	x = 1. The function is strictly monotonic:
		f(x) = (p^cx - 1) / (p^c - 1)
	Unless x is close to 0, the function behaves like f(x) = p^cx / p^c
	pseudo_exp2 () et pseudo_exp10 () are specialised functions for
	respectively p = 2 and p = 10.
Input parameters:
	- x: Number to be pseudo-exponentiated. >= 0.
	- p: Exponent base. > 0.
	- c: Function curvature. The smaller c, the more linear the curve looks.
		> 0.
Returns: The function value.
Throws: Nothing
==============================================================================
*/

double	pseudo_exp (double x, double p, double c)
{
	assert (x >= 0);
	assert (p > 0);
	assert (c > 0);

	using namespace std;

	const double	clogp = c * log (p);
	const double	num = exp (clogp * x) - 1;
	const double	den = exp (clogp) - 1;

	return (num / den);
}



double	pseudo_exp2 (double x, double c)
{
	assert (x >= 0);
	assert (c > 0);

	using namespace std;

	const double	clogp = c * LN2;
	const double	num = exp (clogp * x) - 1;
	const double	den = exp (clogp) - 1;

	return (num / den);
}



double	pseudo_exp10 (double x, double c)
{
	assert (x >= 0);
	assert (c > 0);

	using namespace std;

	const double	clogp = c * LN10;
	const double	num = exp (clogp * x) - 1;
	const double	den = exp (clogp) - 1;

	return (num / den);
}



/*
==============================================================================
Name: pseudo_log
Description:
	Inverse function of pseudo_exp(). The real function is:
		g(y) = log (y * (p^c-1) + 1) / (c * log (p))
Input parameters:
	- y: number to be logarithmed. Must be >= 0.
	- p: logarithm base. > 0.
	- c: Curvature. The smaller c, the more linear the curve looks. > 0.
Returns: The function value.
Throws: Nothing
==============================================================================
*/

double	pseudo_log (double y, double p, double c)
{
	assert (y >= 0);
	assert (p > 0);
	assert (c > 0);

	using namespace std;

	const double	clogp = c * log (p);
	const double	num = log (y * (exp (clogp) - 1) + 1);

	return (num / clogp);
}



double	pseudo_log2 (double y, double c)
{
	assert (y >= 0);
	assert (c > 0);

	using namespace std;

	const double	clogp = c * LN2;
	const double	num = log (y * (exp (clogp) - 1) + 1);

	return (num / clogp);
}



double	pseudo_log10 (double y, double c)
{
	assert (y >= 0);
	assert (c > 0);

	using namespace std;

	const double	clogp = c * LN10;
	const double	num = log (y * (exp (clogp) - 1) + 1);

	return (num / clogp);
}



/*
==============================================================================
Name: get_next_pow2
Description:
	Returns the exponent of the smallest power of 2 equal or greater to the
	parameter. Equivalent to int (ceil (log (x) / log (2))).
Input parameters:
	- x: number whose power of 2 has to be calculated. Must be > 0.
Returns: The exponent of the power of 2.
Throws: Nothing.
==============================================================================
*/

int	get_next_pow2 (long x)
{
	assert (x > 0);

	int		p;
	--x;

#if defined (_MSC_VER)

	if (x == 0)
	{
		p = 0;
	}
	else
	{
#if (archi_WORD_SIZE == 64)

		unsigned long	kk;
		_BitScanReverse (&kk, x);
		p = kk + 1;

#else

		__asm
		{
			xor				eax, eax
			bsr				eax, x
			inc				eax
			mov				p, eax
		}

#endif
	}

#else		// WIN32 && _MSC_VER

	p = 0;
	while ((x & ~0xFFFFL) != 0)
	{
		p += 16;
		x >>= 16;
	}
	while ((x & ~0xFL) != 0)
	{
		p += 4;
		x >>= 4;
	}
	while (x > 0)
	{
		++p;
		x >>= 1;
	}

#endif	// WIN32 && _MSC_VER

	return (p);
}



int	get_prev_pow2 (long x)
{
	assert (x > 0);

	int		p;

#if defined (_MSC_VER)

	if (x != 0)
	{
#if (archi_WORD_SIZE == 64)

		unsigned long	kk;
		_BitScanReverse (&kk, x);
		p = kk;

#else

		__asm
		{
			xor				eax, eax
			bsr				eax, x
			mov				p, eax
		}

#endif
	}

#else		// WIN32 && _MSC_VER

	p = -1;
	while ((x & ~0xFFFFL) != 0)
	{
		p += 16;
		x >>= 16;
	}
	while ((x & ~0xFL) != 0)
	{
		p += 4;
		x >>= 4;
	}
	while (x > 0)
	{
		++p;
		x >>= 1;
	}

#endif	// WIN32 && _MSC_VER

	return (p);
}



bool	is_pow2 (long x)
{
	assert (x > 0);

	return  ((x & -x) == x);
}



long	conv_int_2_bcd (long x)
{
	assert (x >= 0);

	long				mult = 1;
	long				y = 0;
	while (x != 0)
	{
		const int		digit = x % 10;
		x /= 10;
		y += digit * mult;
		mult <<= 4;
	}

	return (y);
}



long	conv_bcd_2_int (long y)
{
	assert (y >= 0);

	long				mult = 1;
	long				x = 0;
	while (y != 0)
	{
		const int		digit = y & 0xF;
		y /= 16;
		x += digit * mult;
		mult *= 10;
	}

	return (x);
}



/*
==============================================================================
Name: conv_db_2_lin_p2
Description:
	Converts a value in approximate dB (+6 = x2) into a linear value, but first
	rounds the input value to 6 dB steps.
Input parameters:
	- x: Number of dB, in [-360 ; +360] range.
Returns: The linear value
Throws: Nothing
==============================================================================
*/

double	conv_db_2_lin_p2 (int x)
{
	assert (x >= -60*6);
	assert (x <=  60*6);

	if (x < -30*6)
	{
		return ((1.0 / (1L << 30)) / (1L << (x/-6 - 30)));
	}
	else if (x < 0)
	{
		return (1.0 / (1L << (x / -6)));
	}
	else if (x > 30*6)
	{
		return (double (1L << 30) * (1L << (x/6 - 30)));
	}

	return (1L << (x / 6));
}



/*
==============================================================================
Name: conv_digit_2_char
Description:
	Convert a value into a digit (character). This digit may be in any base.
	Function to be used to print hexadecimal digits.
Input parameters:
	- x: Number to convert into a character.
Returns: The character (upper case).
Throws: Nothing
==============================================================================
*/

char	conv_digit_2_char (int x)
{
	assert (x >= 0);
	assert (x <= 'Z' - 'A' + 10);

	char				c;
	if (x < 10)
	{
		c = x + '0';
	}
	else
	{
		c = x + 'A' - 10;
	}

	return (c);
}



/*
==============================================================================
Name: conv_char_2_digit
Description:
	Convert a digit or letter into a number. This number is the value of the
	character. Thus 'A' = 10, 'C' = 12 etc. Use this function to analyse
	hexadecimal characters.
Input parameters:
	- c: The character to convert.
Returns: Digit value.
Throws: Nothing
==============================================================================
*/

int	conv_char_2_digit (char c)
{
	using namespace std;

	c = toupper (c);

	assert (   (c >= '0' && c <= '9')
	        || (c >= 'A' && c <= 'Z'));

	int				x;
	if (c >= '0' && c <= '9')
	{
		x = c - '0';
	}
	else
	{
		x = c - 'A' + 10;
	}

	return (x);
}



/*
==============================================================================
Name: change_case
Description: Change the case of a character.
Input parameters:
	- c: The character which case is to change.
	- lower_case_flag: true if the character should be converted to lower case,
		false if it should be converted to upper case.
Returns: The converted character.
Throws: Nothing
==============================================================================
*/

char	change_case (char c, bool lower_case_flag)
{
	using namespace std;

	if (lower_case_flag)
	{
		c = tolower (c);
	}
	else
	{
		c = toupper (c);
	}

	return (c);
}



template <class T>
T	shape_s (T x)
{
	const T		xx = x * x;

	return ((static_cast <T> (3) - static_cast <T> (2) * x) * xx);
}



template <class T>
T	shape_saturate (T x)
{
	return (1 - shape_desaturate (static_cast <T> (1) - x));
}



template <class T>
T	shape_desaturate (T x)
{
	return (x * x);
}



/*
==============================================================================
Name: get_strlen
Description: Just a wrapper for std::strlen ()
Input parameters:
	- text_0: Null-terminated text whose length should be evaluated.
Returns: The string length.
Throws: Nothing
==============================================================================
*/

long	get_strlen (const char *text_0)
{
	using namespace std;

	return (long (strlen (text_0)));
}



/*\\\ PROTECTED \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/



/*\\\ PRIVATE \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/



}	// namespace basic



#endif	// basic_basic_fnc_CODEHEADER_INCLUDED



/*\\\ EOF \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/
