/*****************************************************************************

        Biquad.cpp

--- Legal stuff ---

This program is free software. It comes without any warranty, to
the extent permitted by applicable law. You can redistribute it
and/or modify it under the terms of the Do What The Fuck You Want
To Public License, Version 2, as published by Sam Hocevar. See
http://sam.zoy.org/wtfpl/COPYING for more details.

*Tab=3***********************************************************************/



#if defined (_MSC_VER)
	#pragma warning (1 : 4130 4223 4705 4706)
	#pragma warning (4 : 4355 4786 4800)
#endif



/*\\\ INCLUDE FILES \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/

#include	"basic/basic_def.h"
#include	"basic/basic_fnc.h"
#include	"dsp/Biquad.h"
#include	"dsp/dsp_def.h"

#include	<cassert>
#include	<cmath>



namespace dsp
{



/*\\\ PUBLIC \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/



/*
==============================================================================
Name: ctor
Description:
	Builds the filter, ready to work (transfer function is 1)
Throws: Nothing
==============================================================================
*/

Biquad::Biquad ()
{
	_z_eq_b [0] = 1;
	_z_eq_b [1] = 0;
	_z_eq_b [2] = 0;
	_z_eq_a [0] = 1;
	_z_eq_a [1] = 0;
	_z_eq_a [2] = 0;

	clear_buffers ();
}



/*
==============================================================================
Name: process_block
Description:
	Filters a block of samples. Can work in-place.
Input parameters:
	- src_ptr: Input sample block
	- nbr_spl: Number of samples to process. >= 0
Output parameters:
	- dest_ptr: Output sample block
Throws: Nothing
==============================================================================
*/

void	Biquad::process_block (float *dest_ptr, const float *src_ptr, long nbr_spl)
{
	assert (nbr_spl >= 0);

	if (nbr_spl == 0)
	{
		return;
	}

	// If we're not on a pair boudary, we process a single sample.
	if (_mem_pos != 0)
	{
		*dest_ptr = process_sample (*src_ptr);
		++ src_ptr;
		++ dest_ptr;
		-- nbr_spl;
	}

	if (nbr_spl == 0)
	{
		return;
	}

	long				half_nbr_spl = nbr_spl >> 1;
	long				index = 0;
	if (half_nbr_spl > 0)
	{
		float				mem_x [2];
		float				mem_y [2];
		mem_x [0] = _mem_x [0];
		mem_x [1] = _mem_x [1];
		mem_y [0] = _mem_y [0];
		mem_y [1] = _mem_y [1];

		do
		{

			float				x = src_ptr [index];
			mem_y [1] =   _z_eq_b [0] * x
			            + (  _z_eq_b [1] * mem_x [0]
			               + _z_eq_b [2] * mem_x [1])
			            - (  _z_eq_a [1] * mem_y [0]
			               + _z_eq_a [2] * mem_y [1]);

			mem_x [1] = x;
			dest_ptr [index] = mem_y [1];

			x = src_ptr [index + 1];
			mem_y [0] =   _z_eq_b [0] * x
			            + (  _z_eq_b [1] * mem_x [1]
			               + _z_eq_b [2] * mem_x [0])
			            - (  _z_eq_a [1] * mem_y [1]
			               + _z_eq_a [2] * mem_y [0]);

			mem_x [0] = x;
			dest_ptr [index + 1] = mem_y [0];
			index += 2;

			-- half_nbr_spl;
		}
		while (half_nbr_spl > 0);

		_mem_x [0] = mem_x [0];
		_mem_x [1] = mem_x [1];
		_mem_y [0] = mem_y [0];
		_mem_y [1] = mem_y [1];
	}

	// If number of samples was odd, there is one more to process.
	if ((nbr_spl & 1) > 0)
	{
		dest_ptr [index] = process_sample (src_ptr [index]);
	}
}


// idem mais en s'ajoutant a la sortie
void	Biquad::process_block_mix (float *dest_ptr, const float *src_ptr, long nbr_spl)
{
	assert (nbr_spl >= 0);

	if (nbr_spl == 0)
	{
		return;
	}

	/* Si on est pas sur le debut d'une paire, on traite un sample normalement
		pour se retrouver sur le debut d'une paire. */
	if (_mem_pos != 0)
	{
		*dest_ptr += process_sample (*src_ptr);
		++ src_ptr;
		++ dest_ptr;
		-- nbr_spl;
	}

	if (nbr_spl == 0)
	{
		return;
	}

	long				half_nbr_spl = nbr_spl >> 1;
	long				index = 0;
	if (half_nbr_spl > 0)
	{
		float				mem_x [2];
		float				mem_y [2];
		mem_x [0] = _mem_x [0];
		mem_x [1] = _mem_x [1];
		mem_y [0] = _mem_y [0];
		mem_y [1] = _mem_y [1];

		do
		{

			float				x = src_ptr [index];
			mem_y [1] =   _z_eq_b [0] * x
			            + (  _z_eq_b [1] * mem_x [0]
			               + _z_eq_b [2] * mem_x [1])
			            - (  _z_eq_a [1] * mem_y [0]
			               + _z_eq_a [2] * mem_y [1]);

			mem_x [1] = x;
			dest_ptr [index] += mem_y [1];

			x = src_ptr [index + 1];
			mem_y [0] =   _z_eq_b [0] * x
			            + (  _z_eq_b [1] * mem_x [1]
			               + _z_eq_b [2] * mem_x [0])
			            - (  _z_eq_a [1] * mem_y [1]
			               + _z_eq_a [2] * mem_y [0]);

			mem_x [0] = x;
			dest_ptr [index + 1] += mem_y [0];
			index += 2;

			-- half_nbr_spl;
		}
		while (half_nbr_spl > 0);

		_mem_x [0] = mem_x [0];
		_mem_x [1] = mem_x [1];
		_mem_y [0] = mem_y [0];
		_mem_y [1] = mem_y [1];
	}

	/* Si le nombre de samples etait impair, il en reste encore un a traiter. */
	if ((nbr_spl & 1) > 0)
	{
		dest_ptr [index] += process_sample (src_ptr [index]);
	}
}



void	Biquad::process_block_nd (float *dest_ptr, const float *src_ptr, long nbr_spl)
{
	assert (nbr_spl >= 0);

	if (nbr_spl == 0)
	{
		return;
	}

	/* Si on est pas sur le debut d'une paire, on traite un sample normalement
		pour se retrouver sur le debut d'une paire. */
	if (_mem_pos != 0)
	{
		*dest_ptr = process_sample (*src_ptr);
		++ src_ptr;
		++ dest_ptr;
		-- nbr_spl;
	}

	if (nbr_spl == 0)
	{
		return;
	}

	long				half_nbr_spl = nbr_spl >> 1;
	long				index = 0;
	if (half_nbr_spl > 0)
	{
		float				mem_x [2];
		float				mem_y [2];
		mem_x [0] = _mem_x [0];
		mem_x [1] = _mem_x [1];
		mem_y [0] = _mem_y [0];
		mem_y [1] = _mem_y [1];

		do
		{

			float				x = src_ptr [index];
			mem_y [1] =   (  _z_eq_b [0] * x
			               + ANTI_DENORMAL)	// On ne le rajoute qu'une fois sur 2 -> DC + Nyquist
			            + (  _z_eq_b [1] * mem_x [0]
			               + _z_eq_b [2] * mem_x [1])
			            - (  _z_eq_a [1] * mem_y [0]
			               + _z_eq_a [2] * mem_y [1]);

			mem_x [1] = x;
			dest_ptr [index] = mem_y [1];

			x = src_ptr [index + 1];
			mem_y [0] =   _z_eq_b [0] * x
			            + (  _z_eq_b [1] * mem_x [1]
			               + _z_eq_b [2] * mem_x [0])
			            - (  _z_eq_a [1] * mem_y [1]
			               + _z_eq_a [2] * mem_y [0]);

			mem_x [0] = x;
			dest_ptr [index + 1] = mem_y [0];
			index += 2;

			-- half_nbr_spl;
		}
		while (half_nbr_spl > 0);

		_mem_x [0] = mem_x [0];
		_mem_x [1] = mem_x [1];
		_mem_y [0] = mem_y [0];
		_mem_y [1] = mem_y [1];
	}

	/* Si le nombre de samples etait impair, il en reste encore un a traiter. */
	if ((nbr_spl & 1) > 0)
	{
		dest_ptr [index] = process_sample (src_ptr [index]);
	}
}



void	Biquad::process_block_clip (float *dest_ptr, const float *src_ptr, long nbr_spl, float limit)
{
	assert (nbr_spl >= 0);

	if (nbr_spl == 0)
	{
		return;
	}

	/* Si on est pas sur le debut d'une paire, on traite un sample normalement
		pour se retrouver sur le debut d'une paire. */
	if (_mem_pos != 0)
	{
		*dest_ptr = process_sample_clip (*src_ptr, limit);
		++ src_ptr;
		++ dest_ptr;
		-- nbr_spl;
	}

	if (nbr_spl == 0)
	{
		return;
	}

	long				half_nbr_spl = nbr_spl >> 1;
	long				index = 0;
	if (half_nbr_spl > 0)
	{
		const float		limit_mul = 1.0f / (6 * limit*limit);
		float				mem_x [2];
		float				mem_y [2];
		mem_x [0] = _mem_x [0];
		mem_x [1] = _mem_x [1];
		mem_y [0] = _mem_y [0];
		mem_y [1] = _mem_y [1];

		do
		{

			float				x = src_ptr [index];
			mem_y [1] =   _z_eq_b [0] * x
			            + (  _z_eq_b [1] * mem_x [0]
			               + _z_eq_b [2] * mem_x [1])
			            - (  _z_eq_a [1] * mem_y [0]
			               + _z_eq_a [2] * mem_y [1]);
			mem_y [1] *= 1 - mem_y [1]*mem_y [1] * limit_mul;

			mem_x [1] = x;
			dest_ptr [index] = mem_y [1];

			x = src_ptr [index + 1];
			mem_y [0] =   _z_eq_b [0] * x
			            + (  _z_eq_b [1] * mem_x [1]
			               + _z_eq_b [2] * mem_x [0])
			            - (  _z_eq_a [1] * mem_y [1]
			               + _z_eq_a [2] * mem_y [0]);
			mem_y [0] *= 1 - mem_y [0]*mem_y [0] * limit_mul;

			mem_x [0] = x;
			dest_ptr [index + 1] = mem_y [0];
			index += 2;

			-- half_nbr_spl;
		}
		while (half_nbr_spl > 0);

		_mem_x [0] = mem_x [0];
		_mem_x [1] = mem_x [1];
		_mem_y [0] = mem_y [0];
		_mem_y [1] = mem_y [1];
	}

	/* Si le nombre de samples etait impair, il en reste encore un a traiter. */
	if ((nbr_spl & 1) > 0)
	{
		dest_ptr [index] = process_sample_clip (src_ptr [index], limit);
	}
}



void	Biquad::process_block_limit (float *dest_ptr, const float *src_ptr, long nbr_spl, EnvFollowerBase &env_follower, float inv_limit)
{
	assert (dest_ptr != 0);
	assert (src_ptr != 0);
	assert (nbr_spl > 0);
	assert (&env_follower != 0);

	/*** A optimiser ***/

	long				pos = 0;
	do
	{
		dest_ptr [pos] =
			process_sample_limit (src_ptr [pos], env_follower, inv_limit);
		++pos;
	}
	while (pos < nbr_spl);
}



/*
==============================================================================
Name: process_block
Description:
	Filters a block of data, in-place
Input parameters:
	- nbr_spl: Number of samples to process, >= 0
Input/output parameters:
	- spl_ptr: Block of data to be filtered
Throws: Nothing
==============================================================================
*/

void	Biquad::process_block (float *spl_ptr, long nbr_spl)
{
	process_block (spl_ptr, spl_ptr, nbr_spl);
}



void	Biquad::process_block_mix (float *spl_ptr, long nbr_spl)
{
	process_block_mix (spl_ptr, spl_ptr, nbr_spl);
}



void	Biquad::process_block_nd (float *spl_ptr, long nbr_spl)
{
	process_block_nd (spl_ptr, spl_ptr, nbr_spl);
}



void	Biquad::process_block_clip (float *spl_ptr, long nbr_spl, float limit)
{
	process_block_clip (spl_ptr, spl_ptr, nbr_spl, limit);
}



void	Biquad::process_block_limit (float *spl_ptr, long nbr_spl, EnvFollowerBase &env_follower, float inv_limit)
{
	process_block_limit (spl_ptr, spl_ptr, nbr_spl, env_follower, inv_limit);
}



void	Biquad::set_mem_dc (float dc)
{
	_mem_x [0] = dc;
	_mem_x [1] = dc;
}



/*==========================================================================*/
/*      Name: clip_memory                                                   */
/*      Description: Sature la memoire Y. Est utile pour eliminer les       */
/*                   transitoires apres le changement d'un parametre du     */
/*                   filtre.                                                */
/*      Input parameters:                                                   */
/*        - limit : valeur absolue a laquelle on veut limiter la memoire.   */
/*                  Doit etre positive ou nulle.                            */
/*      Throws: Rien                                                        */
/*==========================================================================*/

void	Biquad::clip_memory (float limit)
{
	assert (limit >= 0);

	if (_mem_y [0] > limit)
	{
		_mem_y [0] = limit;
	}
	else if (_mem_y [0] < -limit)
	{
		_mem_y [0] = -limit;
	}

	if (_mem_y [1] > limit)
	{
		_mem_y [1] = limit;
	}
	else if (_mem_y [1] < -limit)
	{
		_mem_y [1] = -limit;
	}
}



/*
==============================================================================
Name: clear_buffers
Description:
	Clears all filter buffers.
Throws: Nothing
==============================================================================
*/

void	Biquad::clear_buffers ()
{
	_mem_x [0] = 0;
	_mem_x [1] = 0;
	_mem_y [0] = 0;
	_mem_y [1] = 0;
	_mem_pos = 0;
}



/*==========================================================================*/
/*      Name: check_denormal                                                */
/*      Description: Verifie qu'il n'y ait pas de nombre denormalise dans   */
/*                   la memoire recursive du filtre. La fonction est faite  */
/*                   pour etre utilisee directement dans un assert ().      */
/*      Returns: true s'il n'y a pas de nombre denormalise.                 */
/*      Throws: Rien                                                        */
/*==========================================================================*/

bool	Biquad::check_denormal () const
{
	return (! (   basic::is_denormal (_mem_y [0])
	           || basic::is_denormal (_mem_y [1])));
}



void	Biquad::kill_denormal ()
{
	if (basic::is_denormal (_mem_y [0]))
	{
		_mem_y [0] = 0;
	}

	if (basic::is_denormal (_mem_y [1]))
	{
		_mem_y [1] = 0;
	}
}



/*\\\ PROTECTED \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/



/*\\\ PRIVATE \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/



}	// namespace dsp



/*\\\ EOF \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/
