/*****************************************************************************

        Biquad.hpp

--- Legal stuff ---

This program is free software. It comes without any warranty, to
the extent permitted by applicable law. You can redistribute it
and/or modify it under the terms of the Do What The Fuck You Want
To Public License, Version 2, as published by Sam Hocevar. See
http://sam.zoy.org/wtfpl/COPYING for more details.

*Tab=3***********************************************************************/



#if ! defined (dsp_Biquad_CODEHEADER_INCLUDED)
#define	dsp_Biquad_CODEHEADER_INCLUDED



/*\\\ INCLUDE FILES \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/

#include	"dsp/dsp_def.h"
#include	"dsp/EnvFollowerBase.h"

#include <cassert>



namespace dsp
{



/*\\\ PUBLIC \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/



/*
==============================================================================
Name: set_z_eq
Description:
	Assigns filter coefficients. Filter equation is:

        2                 2
       ---               ---
y[t] = \   b[k].x[t-k] - \   a[k].y[t_k]
       /__               /__
       k=0               k=1

Input parameters:
	- b: Direct coefficient part.
	- a: Recursive coefficient part. 0-order coefficient is not taken into
		account.
Throws: Nothing
==============================================================================
*/

void	Biquad::set_z_eq (const float b [3], const float a [3])
{
	_z_eq_b [0] = b [0];
	_z_eq_b [1] = b [1];
	_z_eq_b [2] = b [2];
	_z_eq_a [1] = a [1];
	_z_eq_a [2] = a [2];
}



void	Biquad::set_mem_x (int delay, float val)
{
	assert (delay >= 0);
	assert (delay < 2);

	_mem_x [delay] = val;
}



float	Biquad::get_mem_x (int delay)
{
	assert (delay >= 0);
	assert (delay < 2);

	return (_mem_x [delay]);
}



void	Biquad::set_mem_y (int delay, float val)
{
	assert (delay >= 0);
	assert (delay < 2);

	_mem_y [delay] = val;
}



float	Biquad::get_mem_y (int delay)
{
	assert (delay >= 0);
	assert (delay < 2);

	return (_mem_y [delay]);
}



/*
==============================================================================
Name: apply_volume
Description:
	Modifies the coefficients in order to modify the filter volume. This
	function should be called only after coefficient calculation. Any new
	coefficient calculation will cancel the volume change.
Input parameters:
	- gain: Gain to apply to the filter. Can be 0.
Throws: Nothing
==============================================================================
*/

void	Biquad::apply_volume (float gain)
{
	_z_eq_b [0] *= gain;
	_z_eq_b [1] *= gain;
	_z_eq_b [2] *= gain;
}



/*
==============================================================================
Name: copy_filter
Description:
	Copies filter characteristics from another filter, without its state.
Input parameters:
	- other: Source biquad.
Throws: Nothing
==============================================================================
*/

void	Biquad::copy_filter (const Biquad &other)
{
	_z_eq_b [0] = other._z_eq_b [0];
	_z_eq_b [1] = other._z_eq_b [1];
	_z_eq_b [2] = other._z_eq_b [2];
	_z_eq_a [1] = other._z_eq_a [1];
	_z_eq_a [2] = other._z_eq_a [2];
}



/*
==============================================================================
Name: process_sample
Description:
	Filters a single sample
Input parameters:
	- x: input sample
Returns: output sample
Throws: Nothing
==============================================================================
*/

float	Biquad::process_sample (float x)
{
	const int		alt_pos = 1 - _mem_pos;
	const float		y =   _z_eq_b [0] * x
						    + (  _z_eq_b [1] * _mem_x [_mem_pos]
						       + _z_eq_b [2] * _mem_x [alt_pos])
						    - (  _z_eq_a [1] * _mem_y [_mem_pos]
						       + _z_eq_a [2] * _mem_y [alt_pos]);

	_mem_x [alt_pos] = x;
	_mem_y [alt_pos] = y;
	_mem_pos = alt_pos;

	return (y);
}



float	Biquad::process_sample_nd (float x)
{
	const int		alt_pos = 1 - _mem_pos;
	const float		y =   (  _z_eq_b [0] * x
						       + ANTI_DENORMAL)
						    + (  _z_eq_b [1] * _mem_x [_mem_pos]
						       + _z_eq_b [2] * _mem_x [alt_pos])
						    - (  _z_eq_a [1] * _mem_y [_mem_pos]
						       + _z_eq_a [2] * _mem_y [alt_pos]);

	_mem_x [alt_pos] = x;
	_mem_y [alt_pos] = y;
	_mem_pos = alt_pos;

	return (y);
}



float	Biquad::process_sample_clip (float x, float limit)
{
	const int		alt_pos = 1 - _mem_pos;
	_mem_y [alt_pos] =   _z_eq_b [0] * x
						    + (  _z_eq_b [1] * _mem_x [_mem_pos]
						       + _z_eq_b [2] * _mem_x [alt_pos])
						    - (  _z_eq_a [1] * _mem_y [_mem_pos]
						       + _z_eq_a [2] * _mem_y [alt_pos]);

	_mem_x [alt_pos] = x;
	_mem_pos = alt_pos;
	clip_memory_fast (limit);

	return (_mem_y [_mem_pos]);
}



float	Biquad::process_sample_limit (float x, EnvFollowerBase &env_follower, float inv_limit)
{
	assert (&env_follower != 0);

	const int		alt_pos = 1 - _mem_pos;
	float				y =   _z_eq_b [0] * x
						    + (  _z_eq_b [1] * _mem_x [_mem_pos]
						       + _z_eq_b [2] * _mem_x [alt_pos])
						    - (  _z_eq_a [1] * _mem_y [_mem_pos]
						       + _z_eq_a [2] * _mem_y [alt_pos]);

	_mem_x [alt_pos] = x;

	const float		vol = env_follower.process_sample (y * inv_limit);
	if (vol > 1)
	{
		y /= vol;
	}
	_mem_y [alt_pos] = y;
	_mem_pos = alt_pos;

	return (y);
}



/*
==============================================================================
Name: clip_memory_fast
Description:
	Like clip_memory(), but a bit faster (and approximated). Efficient only
	if the magnitude of the value to be clipped is below 2.5 * limit.
Input parameters:
	- limit: Maximum value. >= 0.
Throws: Nothing
==============================================================================
*/

void	Biquad::clip_memory_fast (float limit)
{
	assert (limit > 0);

	const float		y = _mem_y [_mem_pos];

	_mem_y [_mem_pos] *= 1 - y*y * (1.0f / (6 * limit*limit));
}



/*\\\ PROTECTED \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/



/*\\\ PRIVATE \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/



}	// namespace dsp



#endif	// dsp_Biquad_CODEHEADER_INCLUDED



/*\\\ EOF \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/
