/*****************************************************************************

        BiquadS.cpp

#* LDS : Probleme des basses frequences

	L'arrondi en float nuit a la precision des filtres en basse frequence,
surtout quand ils font partie d'une cascade d'ordre eleve. En effet la
mantisse de 23 bits est souvent trop faible pour contenir des coefficients
de type 1+x avec x petit. De plus les erreurs d'arrondis peuvent s'accumuler
dans la memoire de maniere significative au cours du filtrage.

	Une solution serait d'offrir une optimisation optionnelle. Les coefficients
seraient calcules a leur precision maximale, en double par exemple. Lors d'une
deuxieme etape ils seraient convertis en float. Cette operation consiste a
reduire la precision de la mantisse en faisant un arrondi. Chaque coefficient
serait ainsi decline en plusieurs versions : arrondi par defaut et par exces.
Il sera peut-etre aussi souhaitable de tester des erreurs d'arrondi excessives
pour etendre le champ d'investigation.

	En regroupant ces coefficients, on obtiendra ainsi un certain nombre de
combinaisons possibles pour un meme filtre. Il reste a evaluer ces
possibilites et a choisir la meilleure. Deux possibilites pour cette
evaluation :

	- Calcul des positions des poles/zeros et comparaison avec les poles/zeros
trouves avec une haute precision. On gardera la solution qui s'eloigne le
moins des valeurs theoriques.

	- Test des filtres en situation reelle et comparaison de leurs reponses
avec celle du filtre theorique. Cette methode est optimale puisqu'elle tient
compte des erreurs d'arrondi survenant au cours du filtrage, mais a
l'inconvenient d'etre tres chere en calcul.

--- Legal stuff ---

This program is free software. It comes without any warranty, to
the extent permitted by applicable law. You can redistribute it
and/or modify it under the terms of the Do What The Fuck You Want
To Public License, Version 2, as published by Sam Hocevar. See
http://sam.zoy.org/wtfpl/COPYING for more details.

*Tab=3***********************************************************************/



#if defined (_MSC_VER)
	#pragma warning (1 : 4130 4223 4705 4706)
	#pragma warning (4 : 4355 4786 4800)
#endif



/*\\\ INCLUDE FILES \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/

#include	"basic/basic_def.h"
#include	"basic/basic_fnc.h"
#include	"dsp/BiquadS.h"

#include	<cmath>



namespace std { }

namespace dsp
{



/*\\\ PUBLIC \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/



BiquadS::BiquadS ()
:	_sample_freq (44100)
,	_f0 (1000)
{
	_s_eq_a [0] = 1;
	_s_eq_a [1] = 2;
	_s_eq_a [2] = 1;

	_s_eq_b [0] = _s_eq_a [0];
	_s_eq_b [1] = _s_eq_a [1];
	_s_eq_b [2] = _s_eq_a [2];
}



/*
==============================================================================
Name: transform_s_to_z
Description:
	Transforms the continous s-plane equation in a discrete z-plane equation
	which can be used directly to filter signal, using the bilinear transfrom.

        ---        k                  ---        -k
        \   a[k].s                    \   a'[k].z
        /__                           /__
H(s) = --------------   ==>   H(z) = ---------------
        ---       k                   ---        -k
        \   b[k].s                    \   b'[k].z
        /__                           /__

Throws: Nothing
==============================================================================
*/

void	BiquadS::transform_s_to_z ()
{
	using namespace std;
	using basic::is_null;
	using basic::PI;

	// s to z bilinear transform
	const double	inv_k = tan (_f0 * PI / _sample_freq);
	assert (! is_null (inv_k));
	const double	k = 1 / inv_k;
	const double	kk = k*k;

	const double	b1k = _s_eq_b [1] * k;
	const double	b2kk = _s_eq_b [2] * kk;
	const double	b2kk_plus_b0 = b2kk + _s_eq_b [0];
	const double	b0z = b2kk_plus_b0 + b1k;
	const double	b2z = b2kk_plus_b0 - b1k;
	const double	b1z = 2 * (_s_eq_b [0] - b2kk);

	const double	a1k = _s_eq_a [1] * k;
	const double	a2kk = _s_eq_a [2] * kk;
	const double	a2kk_plus_a0 = a2kk + _s_eq_a [0];
	const double	a0z = a2kk_plus_a0 + a1k;
	const double	a2z = a2kk_plus_a0 - a1k;
	const double	a1z = 2 * (_s_eq_a [0] - a2kk);

	// IIR coefficients
	assert (! is_null (a0z));
	const double	mult = 1 / a0z;

	_z_eq_b [0] = float (b0z * mult);
	_z_eq_b [1] = float (b1z * mult);
	_z_eq_b [2] = float (b2z * mult);

	_z_eq_a [0] = 1;
	_z_eq_a [1] = float (a1z * mult);
	_z_eq_a [2] = float (a2z * mult);
}



/*
==============================================================================
Name : transform_s_to_z_fast
Description :
	Same as transform_s_to_z, but use taylor approximation of tan() to compute
	coefficients. It is less accurate when cutoff frequency is close to
	Nyquist frequency. Also use 32-bit float for coefficient calculation.
Throws : Nothing
==============================================================================
*/

void	BiquadS::transform_s_to_z_fast ()
{
	using basic::is_null;
	using basic::PI;

	// s to z bilinear transform
	const float		inv_k =
		basic::tan_taylor_5th (static_cast <float> (_f0 * PI / _sample_freq));
	assert (! is_null (inv_k));
	const float		k = 1 / inv_k;
	const float		kk = k*k;

	const float		b1k = _s_eq_b [1] * k;
	const float		b2kk = _s_eq_b [2] * kk;
	const float		b2kk_plus_b0 = b2kk + _s_eq_b [0];
	const float		b0z = b2kk_plus_b0 + b1k;
	const float		b2z = b2kk_plus_b0 - b1k;
	const float		b1z = 2 * (_s_eq_b [0] - b2kk);

	const float		a1k = _s_eq_a [1] * k;
	const float		a2kk = _s_eq_a [2] * kk;
	const float		a2kk_plus_a0 = a2kk + _s_eq_a [0];
	const float		a0z = a2kk_plus_a0 + a1k;
	const float		a2z = a2kk_plus_a0 - a1k;
	const float		a1z = 2 * (_s_eq_a [0] - a2kk);

	// IIR coefficients
	assert (! is_null (a0z));
	const float	mult = 1 / a0z;

	_z_eq_b [0] = b0z * mult;
	_z_eq_b [1] = b1z * mult;
	_z_eq_b [2] = b2z * mult;

	_z_eq_a [0] = 1;
	_z_eq_a [1] = a1z * mult;
	_z_eq_a [2] = a2z * mult;
}



float	BiquadS::wrap_freq_blt (float f) const
{
	assert (f > 0);
	assert (f < _sample_freq);
	assert (_f0 < _sample_freq);

	using namespace std;

	const float		mult = static_cast <float> (basic::PI / _sample_freq);
	const float		wrapped_f = float (tan (mult * f) / tan (mult * _f0));

	return (wrapped_f);
}



float	BiquadS::wrap_freq_blt_fast (float f) const
{
	assert (f > 0);
	assert (f < _sample_freq);
	assert (_f0 < _sample_freq);

	using namespace std;

	const float		mult = static_cast <float> (basic::PI / _sample_freq);
	const float		wrapped_f =
		basic::tan_taylor_5th (mult * f) / basic::tan_taylor_5th (mult * _f0);

	return (wrapped_f);
}



/*\\\ PROTECTED \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/



/*\\\ PRIVATE \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/



}	// namespace dsp



/*\\\ EOF \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/
