/*****************************************************************************

        InterpFncPeriodic.hpp

--- Legal stuff ---

This program is free software. It comes without any warranty, to
the extent permitted by applicable law. You can redistribute it
and/or modify it under the terms of the Do What The Fuck You Want
To Public License, Version 2, as published by Sam Hocevar. See
http://sam.zoy.org/wtfpl/COPYING for more details.

*Tab=3***********************************************************************/



#if ! defined (dsp_InterpFncPeriodic_CODEHEADER_INCLUDED)
#define	dsp_InterpFncPeriodic_CODEHEADER_INCLUDED



/*\\\ INCLUDE FILES \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/

#include	"archi/archi_def.h"
#include	"basic/basic_fnc.h"
#include	"dsp/dsp_def.h"

#include	<cmath>



namespace dsp
{



/*\\\ PUBLIC \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/



/*
==============================================================================
Name: ctor
Throws: Depends on GenFtor::operator ()
==============================================================================
*/

template <int CYCLE_RES_LOG2, class GenFtor>
InterpFncPeriodic <CYCLE_RES_LOG2, GenFtor>::InterpFncPeriodic ()
{
	init_coef ();
}



/*
==============================================================================
Name: init_coef
Description:
	Intialises the coefficients (once). Called at the object construction, but
	can be called before.
Throws: Depends on GenFtor::operator ()
==============================================================================
*/

template <int CYCLE_RES_LOG2, class GenFtor>
void	InterpFncPeriodic <CYCLE_RES_LOG2, GenFtor>::init_coef ()
{
	if (! coef_init_flag)
	{
		double			y [CYCLE_RES + 1];
		double			slope [CYCLE_RES + 1];
		int				x;
		GenFtor			fnc;
		const double	scale = 1.0 / CYCLE_RES;

		for (x = 0; x <= CYCLE_RES; ++x)
		{
			y [x] = double (fnc (x * scale));
			slope [x] =   (  double (fnc ((x + 1.0/2048) * scale))
			               - double (fnc ((x - 1.0/2048) * scale)))
			            * 1024;
		}

		for (x = 0; x < CYCLE_RES; ++x)
		{
			const double	p = slope [x];
			const double	q = slope [x+1];
			const double	k = y [x+1] - y [x];

			// Computes coefs in [0 ; 1]
			const double	A =  q +   p - 2*k;
			const double	B = -q - 2*p + 3*k;
			const double	C = p;
			const double	D = y [x];

			// Scales to [x ; x+1]
			const double	a =    A;
			const double	b = -3*A*x + B;
			const double	c = (3*A*x - B*2) * x + C;
			const double	d = ((-A*x + B)   * x - C) * x + D;

			coef [x] [0] = float (d);
			coef [x] [1] = float (c);
			coef [x] [2] = float (b);
			coef [x] [3] = float (a);
		}

		coef_init_flag = true;
	}
}



/*
==============================================================================
Name: operator
Description:
	Returns the approxiamtion of the function for a given point.
Input parameters:
	- x: point where the function must be evaluated.
Returns: Function value
Throws: Nothing
==============================================================================
*/

template <int CYCLE_RES_LOG2, class GenFtor>
float	InterpFncPeriodic <CYCLE_RES_LOG2, GenFtor>::operator () (float x) const
{

#if defined (WIN32) && defined (_MSC_VER)

	#if (archi_WORD_SIZE == 64)

	int				pos = basic::float_to_int_fast (x - 0.5f);
	x -= (pos & ~CYCLE_MASK);
	pos &= CYCLE_MASK;

	#else

	/***
	Important, le mode d'arrondi par defaut doit etre au plus pres.
	Mettre une option pour verifier ca en mode release.
	***/

	static const float	round = -0.5f;
	int				base_pos = ~CYCLE_MASK;
	int				pos;

	__asm
	{
		fld				x
		fld				st (0)

		fadd				round
		fistp				pos

		mov				eax, pos
		and				base_pos, eax
		fisub				base_pos
		fstp				x

		sub				eax, base_pos
		mov				pos, eax
	}

	#endif

#else

	const double	y = x + 1048576.0;	// Avoid rounding problems in negative range (toward 0)
	int				pos = int (y);
	x = float (y - (pos & (~CYCLE_MASK)));
	pos &= CYCLE_MASK;

#endif

	return (       coef [pos] [0]
	        + x * (coef [pos] [1]
	        + x * (coef [pos] [2]
	        + x *  coef [pos] [3])));
}



/*\\\ PROTECTED \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/



/*\\\ PRIVATE \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/



template <int CYCLE_RES_LOG2, class GenFtor>
float	InterpFncPeriodic <CYCLE_RES_LOG2, GenFtor>::coef [CYCLE_RES] [4];

template <int CYCLE_RES_LOG2, class GenFtor>
bool	InterpFncPeriodic <CYCLE_RES_LOG2, GenFtor>::coef_init_flag = false;



}	// namespace dsp



#endif	// dsp_InterpFncPeriodic_CODEHEADER_INCLUDED



/*\\\ EOF \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/
