/*****************************************************************************

        Lfo.cpp

--- Legal stuff ---

This program is free software. It comes without any warranty, to
the extent permitted by applicable law. You can redistribute it
and/or modify it under the terms of the Do What The Fuck You Want
To Public License, Version 2, as published by Sam Hocevar. See
http://sam.zoy.org/wtfpl/COPYING for more details.

*Tab=3***********************************************************************/



#if defined (_MSC_VER)
	#pragma warning (1 : 4130 4223 4705 4706)
	#pragma warning (4 : 4355 4786 4800)
#endif



/*\\\ INCLUDE FILES \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/

#include	"basic/basic_fnc.h"
#include	"dsp/Lfo.h"

#include	<cassert>



namespace dsp
{



/*\\\ PUBLIC \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/



const char *const	Lfo::WF_TYPE_NAME_0 [Lfo::NBR_WF_TYPES] =
{
	"Sine", "Triangle", "Square", "Ramp Up",
	"Cos Up", "Ramp Down", "Cos Down", "Random",
	"Brown Noise", "Red Noise"
};



/*
==============================================================================
Name: ctor
Throws: Nothing
==============================================================================
*/

Lfo::Lfo ()
:	_type (SINE)
,	_amp (1)
,	_fs (44100)
,	_period (0.5)
{
	clear_buffers ();
}



/*
==============================================================================
Name: tick
Description:
	Moves forward the LFO from a given number of samples.
Input parameters:
	- nbr_spl: Number of samples. Must be >= 0.
Throws: Nothing
==============================================================================
*/

void	Lfo::tick (long nbr_spl)
{
	using namespace std;

	assert (nbr_spl >= 0);

	switch (_type)
	{
	case	RANDOM:
		_pos += _step * nbr_spl;
		if (_pos >= CYCLE_RES)
		{
			_pos = fmod (_pos, CYCLE_RES);
			gen_rand_val ();
		}
		break;

	case	BROWN_NOISE:
	case	RED_NOISE:
		_noise_pos += nbr_spl;
		while (_noise_pos >= NOISE_RES)
		{
			_noise_pos -= NOISE_RES;
			gen_rand_val ();
		}
		break;

	default:
		_pos = fmod (_pos + _step * nbr_spl, CYCLE_RES);
		break;
	}
}



/*
==============================================================================
Name: clear_buffers
Description:
	Resets the LFO memory and phase.
Throws: Nothing
==============================================================================
*/

void	Lfo::clear_buffers ()
{
	_pos = 0;
	_random_val = 0;
	_noise_pos = 0;
	_cur_val_arr [0] = 0;
	_cur_val_arr [1] = 0;
	update_param ();
}



/*\\\ PROTECTED \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/



/*
==============================================================================
Name: update_param
Description:
	Updates the LFO parameters.
Throws: Nothing
==============================================================================
*/

void	Lfo::update_param ()
{
	_step = CYCLE_RES / (_period * _fs);
	if (   _type == BROWN_NOISE
	    || _type == RED_NOISE)
	{
		using namespace std;

		const float		pseudo_fs = _fs * (1.0f / NOISE_RES);
		const float		pseudo_f0 = basic::min (float (1.0 / _period),
							                        pseudo_fs * 0.4375f);

		const double	freq_norm = pseudo_f0 / pseudo_fs;
		const double	omega = freq_norm * (basic::PI * 2);
		const double	k = 2 - cos (omega);
		_filter_coef = static_cast <float> (1 + sqrt (k*k - 1) - k);

		_filter_amp = float (sqrt (pseudo_fs / (pseudo_f0 * 2)));
	}
}



/*\\\ PRIVATE \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/



}	// namespace dsp



/*\\\ EOF \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/
