/*****************************************************************************

        Lfo.hpp

--- Legal stuff ---

This program is free software. It comes without any warranty, to
the extent permitted by applicable law. You can redistribute it
and/or modify it under the terms of the Do What The Fuck You Want
To Public License, Version 2, as published by Sam Hocevar. See
http://sam.zoy.org/wtfpl/COPYING for more details.

*Tab=3***********************************************************************/



#if ! defined (dsp_Lfo_CODEHEADER_INCLUDED)
#define	dsp_Lfo_CODEHEADER_INCLUDED



/*\\\ INCLUDE FILES \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/

#include	<cassert>
#include	<cmath>



namespace std { }



namespace dsp
{



/*\\\ PUBLIC \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/



/*
==============================================================================
Name: set_sample_freq
Description:
	Sets the sampling frequency.
Input parameters:
	- fs: Sampling frequency, in Hz, > 0.
Throws: Nothing
==============================================================================
*/

void	Lfo::set_sample_freq (float fs)
{
	assert (fs > 0);
	assert (_period >= 2 / fs);

	_fs = fs;
	update_param ();
}



float	Lfo::get_sample_freq () const
{
	return (_fs);
}



/*
==============================================================================
Name: set_type
Description:
	Sets the LFO type. Changing the LFO type can cause a discontinuity in
	the generated signal.
Input parameters:
	- type: the LFO type (see enum)
Throws: Nothing
==============================================================================
*/

void	Lfo::set_type (WfType type)
{
	_type = type;
	update_param ();
}



/*
==============================================================================
Name: set_period
Description:
	Sets the LFO period.
Input parameters:
	- period: period, in seconds. Must be >= 2/Fs.
Throws: Nothing
==============================================================================
*/

void	Lfo::set_period (double period)
{
	assert (period >= 2 / _fs);

	_period = period;
	update_param ();
}



double	Lfo::get_period () const
{
	return (_period);
}



/*
==============================================================================
Name: set_amp
Description:
	Sets the LFO amplitude.
Input parameters:
	- amp: The amplitude. Can be negative or 0.
Throws: Nothing
==============================================================================
*/

void	Lfo::set_amp (float amp)
{
	_amp = amp;
}



float	Lfo::get_amp () const
{
	return (_amp);
}



/*
==============================================================================
Name: set_phase
Description:
	Sets the normalised LFO phase.
Input parameters:
	- phase: Normalised phase, in [0 ; 1[ range.
Throws: Nothing
==============================================================================
*/

void	Lfo::set_phase (float phase)
{
	assert (phase >= 0);
	assert (phase < 1);

	_pos = phase * CYCLE_RES;
}



float	Lfo::get_phase () const
{
	return (float (_pos * (1.0 / CYCLE_RES)));
}



/*
==============================================================================
Name: get_value
Description:
	Returns the current LFO value, modulated by its amplitude.
Returns: The LFO value. Its magnitude may exceed the specified amplitude.
Throws: Nothing.
==============================================================================
*/

float	Lfo::get_value () const
{
	return (get_raw_value () * _amp);
}



/*
==============================================================================
Name: get_raw_value
Description:
	Returns the current raw LFO value, NOT modulated by its amplitude. Its
	nominal magnitude is 1, although the value it may exceed it.
Returns: The raw LFO value.
Throws: Nothing.
==============================================================================
*/

float	Lfo::get_raw_value () const
{
	using namespace std;

	switch (_type)
	{
	case	SINE:
		return (_sin_gen (float (_pos)));
		break;

	case	TRIANGLE:
		return (float (fabs (_pos * (4.0/CYCLE_RES) - 2) - 1));
		break;

	case	SQUARE:
		if (_pos < 0.5 * CYCLE_RES)
		{
			return (1);
		}
		else
		{
			return (-1);
		}
		break;

	case	RAMP_UP:
		return (float (_pos * (2.0/CYCLE_RES) - 1));
		break;

	case	COS_UP:
		return (_sin_gen (float (_pos * 0.5 + CYCLE_RES * 3 / 4)));
		break;

	case	RAMP_DOWN:
		return (float (1 - _pos * (2.0/CYCLE_RES)));
		break;

	case	COS_DOWN:
		return (_sin_gen (float (_pos * 0.5 + CYCLE_RES / 4)));
		break;

	case	RANDOM:
		return (_random_val);
		break;

	case	RED_NOISE:
		return (_cur_val_arr [MAX_FILTER_STAGES - 1]);
		break;

	case	BROWN_NOISE:
		return (_cur_val_arr [0]);
		break;

	default:
		assert (false);
		break;
	}

	return (0);
}



/*\\\ PROTECTED \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/



/*\\\ PRIVATE \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/



/*
==============================================================================
Name : gen_rand_val
Description :
	Generate a random value for "Random", "Brown Noise" and "Red Noise" types.
	Fills _cur_val_arr [] and _random_val. This latter values is in the range
	[-1 ; 1]. _cur_val_arr [] values can be slightly outside.
Throws : Nothing.
==============================================================================
*/

void	Lfo::gen_rand_val ()
{
	const long		h_m_val = (FastRand::M_VAL - 1) / 2 + 1;
	assert (h_m_val > 0);
	_random_val =   static_cast <float> (_rand_gen.get_new_val () - h_m_val)
	              * (1.0f / h_m_val);

	assert (MAX_FILTER_STAGES == 2);
	const float		amp_rand_val = _random_val * _filter_amp;
	_cur_val_arr [0] += _filter_coef * (amp_rand_val - _cur_val_arr [0]);
	_cur_val_arr [1] += _filter_coef * (_cur_val_arr [0] - _cur_val_arr [1]);
}



}	// namespace dsp



#endif	// dsp_Lfo_CODEHEADER_INCLUDED



/*\\\ EOF \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/
