/*****************************************************************************

        OnePole.cpp

--- Legal stuff ---

This program is free software. It comes without any warranty, to
the extent permitted by applicable law. You can redistribute it
and/or modify it under the terms of the Do What The Fuck You Want
To Public License, Version 2, as published by Sam Hocevar. See
http://sam.zoy.org/wtfpl/COPYING for more details.

*Tab=3***********************************************************************/



#if defined (_MSC_VER)
	#pragma warning (1 : 4130 4223 4705 4706)
	#pragma warning (4 : 4355 4786 4800)
#endif



/*\\\ INCLUDE FILES \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/

#include	"basic/basic_def.h"
#include	"basic/basic_fnc.h"
#include	"dsp/OnePole.h"

#include	<cassert>
#include	<cmath>



namespace dsp
{



/*\\\ PUBLIC \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/



/*
==============================================================================
Name: ctor
Description:
	Initialises the filter with a neutral transfer function.
Throws: Nothing
==============================================================================
*/

OnePole::OnePole ()
{
	_z_eq_b [0] = 1;
	_z_eq_b [1] = 0;
	_z_eq_a [0] = 1;
	_z_eq_a [1] = 0;

	clear_buffers ();
}



/*
==============================================================================
Name: process_block
Description:
	Applies the filter on a block of samples. This function can work "in
	place".
Input parameters:
	- src_ptr: Samples to filter
	- nbr_spl: Number of samples to process. >= 0.
Output parameters:
	- dest_ptr: Filtered sample block.
Throws: Nothing
==============================================================================
*/

void	OnePole::process_block (float *dest_ptr, const float *src_ptr, long nbr_spl)
{
	assert (nbr_spl >= 0);

	if (nbr_spl == 0)
	{
		return;
	}

	float				mem_x = _mem_x;
	float				y = _mem_y;

	do
	{
		const float		x = *src_ptr;
		y =   _z_eq_b [0] * x
		    + _z_eq_b [1] * mem_x
		    - _z_eq_a [1] * y;

		*dest_ptr = y;
		mem_x = x;

		++ src_ptr;
		++ dest_ptr;
		-- nbr_spl;
	}
	while (nbr_spl > 0);

	_mem_x = mem_x;
	_mem_y = y;
}



void	OnePole::process_block_nd (float *dest_ptr, const float *src_ptr, long nbr_spl)
{
	assert (nbr_spl >= 0);

	if (nbr_spl == 0)
	{
		return;
	}

	float				mem_x = _mem_x;
	float				y = _mem_y;

	do
	{
		const float		x = *src_ptr;
		y =   _z_eq_b [0] * x
		    + _z_eq_b [1] * mem_x
		    - (  _z_eq_a [1] * y
		       - ANTI_DENORMAL);

		*dest_ptr = y;
		mem_x = x;

		++ src_ptr;
		++ dest_ptr;
		-- nbr_spl;
	}
	while (nbr_spl > 0);

	_mem_x = mem_x;
	_mem_y = y;
}



void	OnePole::set_mem_dc (float dc)
{
	_mem_x = dc;
}



/*
==============================================================================
Name: clip_memory
Description:
	Saturates the Y (feedback) memory. Can be used to dim sharp transcients
	after an abrupt parameter change.
Input parameters:
	- limit: Maximum value. >= 0.
Throws: Nothing
==============================================================================
*/

void	OnePole::clip_memory (float limit)
{
	assert (limit >= 0);

	if (_mem_y > limit)
	{
		_mem_y = limit;
	}
	else if (_mem_y < -limit)
	{
		_mem_y = -limit;
	}
}



/*
==============================================================================
Name: clear_buffers
Description:
	Clears the filter history.
Throws: Nothing
==============================================================================
*/

void	OnePole::clear_buffers ()
{
	_mem_x = 0;
	_mem_y = 0;
}



/*
==============================================================================
Name: check_denormal
Description:
	Checks if there is not denormal number in the filter memory. Can be used
	directly in an assert().
Returns: true if there is no denormal number.
Throws: Nothing
==============================================================================
*/

bool	OnePole::check_denormal () const
{
	return (! basic::is_denormal (_mem_y));
}



void	OnePole::kill_denormal ()
{
	if (basic::is_denormal (_mem_y))
	{
		_mem_y = 0;
	}
}



/*\\\ PROTECTED \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/



/*\\\ PRIVATE \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/



}	// namespace dsp



/*\\\ EOF \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/
