/*****************************************************************************

        OnePole.hpp

--- Legal stuff ---

This program is free software. It comes without any warranty, to
the extent permitted by applicable law. You can redistribute it
and/or modify it under the terms of the Do What The Fuck You Want
To Public License, Version 2, as published by Sam Hocevar. See
http://sam.zoy.org/wtfpl/COPYING for more details.

*Tab=3***********************************************************************/



#if ! defined (dsp_OnePole_CODEHEADER_INCLUDED)
#define	dsp_OnePole_CODEHEADER_INCLUDED



/*\\\ INCLUDE FILES \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/

#include	"dsp/dsp_def.h"

#include	<cassert>



namespace dsp
{



/*\\\ PUBLIC \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/



/*
==============================================================================
Name: set_z_eq
Description:
	Assigns the filter coefficients.

		y[t] = b[0].x[t-0] + b[1].x[t-1] - a[1].y[t-1]

Input parameters:
	- b: Direct vector
	- a: Feedback vector
Throws: Nothing
==============================================================================
*/

void	OnePole::set_z_eq (const float b [2], const float a [2])
{
	_z_eq_b [0] = b [0];
	_z_eq_b [1] = b [1];
	_z_eq_a [1] = a [1];
}



/*
==============================================================================
Name: apply_volume
Description:
	Modifies the coefficients in order to modify the filter volume. This
	function should be called only after coefficient calculation. Any new
	coefficient calculation will cancel the volume change.
Input parameters:
	- gain: Gain to apply to the filter. Can be 0.
Throws: Nothing
==============================================================================
*/

void	OnePole::apply_volume (float gain)
{
	_z_eq_b [0] *= gain;
	_z_eq_b [1] *= gain;
}



/*
==============================================================================
Name: copy_filter
Description:
	Copies the filter, without its state.
Input parameters:
	- other: The source filter.
Throws: Nothing
==============================================================================
*/

void	OnePole::copy_filter (const OnePole &other)
{
	_z_eq_b [0] = other._z_eq_b [0];
	_z_eq_b [1] = other._z_eq_b [1];
	_z_eq_a [1] = other._z_eq_a [1];
}



/*
==============================================================================
Name: process_sample
Description:
	Filters a single sample.
Input parameters:
	- x: The sample to process.
Returns: The processed sample.
Throws: Nothing
==============================================================================
*/

float	OnePole::process_sample (float x)
{
	const float		y =   _z_eq_b [0] * x
						    + _z_eq_b [1] * _mem_x
						    - _z_eq_a [1] * _mem_y;

	_mem_x = x;
	_mem_y = y;

	return (y);
}



float	OnePole::process_sample_nd (float x)
{
	const float		y =   _z_eq_b [0] * x
						    + _z_eq_b [1] * _mem_x
						    - (  _z_eq_a [1] * _mem_y
						       - ANTI_DENORMAL);

	_mem_x = x;
	_mem_y = y;

	return (y);
}



/*
==============================================================================
Name: process_block
Description:
	Applies the filter on a block of samples, "in place".
Input parameters:
	- nbr_spl: Number of samples to process. >= 0.
Input/output parameters:
	- spl_ptr: Block of samples to be filtered.
Throws: Nothing
==============================================================================
*/

void	OnePole::process_block (float *spl_ptr, long nbr_spl)
{
	process_block (spl_ptr, spl_ptr, nbr_spl);
}



void	OnePole::process_block_nd (float *spl_ptr, long nbr_spl)
{
	process_block_nd (spl_ptr, spl_ptr, nbr_spl);
}



/*
==============================================================================
Name: clip_memory_fast
Description:
	Like clip_memory(), but a bit faster (and approximated).
Input parameters:
	- limit: Maximum value. >= 0.
Throws: Nothing
==============================================================================
*/

void	OnePole::clip_memory_fast (float limit)
{
	assert (limit > 0);

	const float		y = _mem_y;

	_mem_y *= 1 - y*y * (1.0f / (6 * limit*limit));
}



float	OnePole::get_latest_output_sample () const
{
	return (_mem_y);
}



/*\\\ PROTECTED \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/



/*\\\ PRIVATE \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/



}	// namespace dsp



#endif	// dsp_OnePole_CODEHEADER_INCLUDED



/*\\\ EOF \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/
