/*****************************************************************************

        ParamInertia.cpp

--- Legal stuff ---

This program is free software. It comes without any warranty, to
the extent permitted by applicable law. You can redistribute it
and/or modify it under the terms of the Do What The Fuck You Want
To Public License, Version 2, as published by Sam Hocevar. See
http://sam.zoy.org/wtfpl/COPYING for more details.

*Tab=3***********************************************************************/



#if defined (_MSC_VER)
	#pragma warning (1 : 4130 4223 4705 4706)
	#pragma warning (4 : 4355 4786 4800)
#endif



/*\\\ INCLUDE FILES \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/

#include	"dsp/ParamInertia.h"

#include	<cassert>



namespace dsp
{



/*\\\ PUBLIC \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/



/*
==============================================================================
Name: ctor
Description:
	Initialises the object with the current value set to 0 and the inertia to
	100 ms.
Throws: Nothing
==============================================================================
*/

ParamInertia::ParamInertia ()
:	_inertia_time (0.100f)
,	_fs (44100)
,	_old_val (0)
,	_new_val (0)
,	_cur_val (0)
,	_step (0)
{
}



/*
==============================================================================
Name: set_sample_freq
Description:
	Sets the sampling frequency.
Input parameters:
	- fs: Sample frequency in Hz, > 0.
Throws: Nothing
==============================================================================
*/

void	ParamInertia::set_sample_freq (float fs)
{
	assert (fs > 0);

	_step *= _fs / fs;
	_fs = fs;
}



/*
==============================================================================
Name: set_val
Description:
	Sets the value to be reached. The current parameter value will ramp
	linearly to this value and will reach the specified value only after the
	ramping time is elapsed.
Input parameters:
	- val: Value to reach
Throws: Nothing
==============================================================================
*/

void	ParamInertia::set_val (float val)
{
	if (_inertia_time == 0)
	{
		force_val (val);
	}

	else
	{
		_old_val = float (_cur_val);
		_new_val = val;
		_step = (_new_val - _old_val) / (_fs * _inertia_time);
	}
}



/*
==============================================================================
Name: tick
Description:
	Skips the given number of samples. If the parameter is ramping, value is
	changed accordingly.
Input parameters:
	- nbr_spl: Number of samples to skip. > 0.
Throws: Nothing
==============================================================================
*/

void	ParamInertia::tick (long nbr_spl)
{
	assert (nbr_spl >= 0);

	if (_step != 0)
	{
		_cur_val += _step * nbr_spl;

		// On a atteint ou depasse ?
		if (_step * (_cur_val - _new_val) >= 0)
		{
			_cur_val = _new_val;
			_old_val = _new_val;
			_step = 0;
		}
	}
}



/*\\\ PROTECTED \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/



/*\\\ PRIVATE \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/



}	// namespace dsp



/*\\\ EOF \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/
