/*****************************************************************************

        ParamInertia.hpp

--- Legal stuff ---

This program is free software. It comes without any warranty, to
the extent permitted by applicable law. You can redistribute it
and/or modify it under the terms of the Do What The Fuck You Want
To Public License, Version 2, as published by Sam Hocevar. See
http://sam.zoy.org/wtfpl/COPYING for more details.

*Tab=3***********************************************************************/



#if ! defined (dsp_ParamInertia_CODEHEADER_INCLUDED)
#define	dsp_ParamInertia_CODEHEADER_INCLUDED



/*\\\ INCLUDE FILES \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/

#include	<cassert>



namespace dsp
{



/*\\\ PUBLIC \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/



/*
==============================================================================
Name: set_inertia_time
Description:
	Sets the inertia time.
Input parameters:
	- inertia_time: Inertia time, in seconds. >= 0. 0 means there is no
		inertia, all changes are instantaneous.
Throws: Nothing
==============================================================================
*/

void	ParamInertia::set_inertia_time (float inertia_time)
{
	assert (inertia_time >= 0);

	_inertia_time = inertia_time;
}



float	ParamInertia::get_inertia_time () const
{
	return (_inertia_time);
}



/*
==============================================================================
Name: force_val
Description:
	Sets the parameter value. Change is immediate, without any ramping.
Input parameters:
	- val: Parameter value to be set. Range is not specified.
Throws: Nothing
==============================================================================
*/

void	ParamInertia::force_val (float val)
{
	_old_val = val;
	_new_val = val;
	_cur_val = val;
	_step = 0;
}



/*
==============================================================================
Name: get_val
Description:
	Returns the current parameter value.
Returns: Value. Range not specified.
Throws: Nothing.
==============================================================================
*/

float	ParamInertia::get_val () const
{
	return (float (_cur_val));
}



/*
==============================================================================
Name: get_target_val
Description:
	Returns the value it should reach when the ramp is over. If the parameter
	is constant, this is the same value as the current one.
Returns: The final value.
Throws: Nothing.
==============================================================================
*/

float	ParamInertia::get_target_val () const
{
	return (float (_new_val));
}



/*
==============================================================================
Name: is_constant
Description:
	Indicates if the parameter is ramping, i.e. after the next call to skip(),
	if the value will change.
Returns: true if it is ramping, false if it is constant.
Throws: Nothing.
==============================================================================
*/

bool	ParamInertia::is_ramping () const
{
	return (_step != 0);
}



void	ParamInertia::stop ()
{
	force_val (static_cast <float> (_cur_val));
}



/*
==============================================================================
Name: clear_buffers
Description:
	Make the parameter reach the steady state. If there was a ramp, the ramp
	is terminated, reaching the target value.
Throws: Nothing.
==============================================================================
*/

void	ParamInertia::clear_buffers ()
{
	force_val (_new_val);
}



/*\\\ PROTECTED \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/



/*\\\ PRIVATE \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/



}	// namespace dsp



#endif	// dsp_ParamInertia_CODEHEADER_INCLUDED



/*\\\ EOF \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\*/
